import autofit as af
from autolens.data.plotters import ccd_plotters
from autolens.lens.plotters import ray_tracing_plotters
from autolens.lens.plotters import lens_fit_plotters
from autolens.model.inversion.plotters import inversion_plotters
from autolens.pipeline.plotters import hyper_plotters


def plot_ccd_for_phase(
    ccd_data,
    mask,
    positions,
    extract_array_from_mask,
    zoom_around_mask,
    units,
    should_plot_as_subplot,
    should_plot_image,
    should_plot_noise_map,
    should_plot_psf,
    should_plot_signal_to_noise_map,
    should_plot_absolute_signal_to_noise_map,
    should_plot_potential_chi_squared_map,
    visualize_path,
    subplot_path,
):

    output_path = af.path_util.make_and_return_path_from_path_and_folder_names(
        path=visualize_path, folder_names=["ccd"]
    )

    if should_plot_as_subplot:

        ccd_plotters.plot_ccd_subplot(
            ccd_data=ccd_data,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            positions=positions,
            units=units,
            output_path=subplot_path,
            output_format="png",
        )

    ccd_plotters.plot_ccd_individual(
        ccd_data=ccd_data,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        should_plot_image=should_plot_image,
        should_plot_noise_map=should_plot_noise_map,
        should_plot_psf=should_plot_psf,
        should_plot_signal_to_noise_map=should_plot_signal_to_noise_map,
        should_plot_absolute_signal_to_noise_map=should_plot_absolute_signal_to_noise_map,
        should_plot_potential_chi_squared_map=should_plot_potential_chi_squared_map,
        units=units,
        output_path=output_path,
        output_format="png",
    )


def plot_ray_tracing_for_phase(
    tracer,
    grid,
    during_analysis,
    mask,
    extract_array_from_mask,
    zoom_around_mask,
    positions,
    units,
    should_plot_as_subplot,
    should_plot_all_at_end_png,
    should_plot_all_at_end_fits,
    should_plot_image_plane_image,
    should_plot_source_plane,
    should_plot_convergence,
    should_plot_potential,
    should_plot_deflections,
    visualize_path,
    subplot_path,
):
    output_path = af.path_util.make_and_return_path_from_path_and_folder_names(
        path=visualize_path, folder_names=["ray_tracing"]
    )

    if should_plot_as_subplot:

        ray_tracing_plotters.plot_ray_tracing_subplot(
            tracer=tracer,
            grid=grid,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            positions=positions,
            units=units,
            output_path=subplot_path,
            output_format="png",
        )

    ray_tracing_plotters.plot_ray_tracing_individual(
        tracer=tracer,
        grid=grid,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        should_plot_profile_image=should_plot_image_plane_image,
        should_plot_source_plane=should_plot_source_plane,
        should_plot_convergence=should_plot_convergence,
        should_plot_potential=should_plot_potential,
        should_plot_deflections=should_plot_deflections,
        units=units,
        output_path=output_path,
        output_format="png",
    )

    if not during_analysis:

        if should_plot_all_at_end_png:

            ray_tracing_plotters.plot_ray_tracing_individual(
                tracer=tracer,
                grid=grid,
                mask=mask,
                extract_array_from_mask=extract_array_from_mask,
                zoom_around_mask=zoom_around_mask,
                positions=positions,
                should_plot_profile_image=True,
                should_plot_source_plane=True,
                should_plot_convergence=True,
                should_plot_potential=True,
                should_plot_deflections=True,
                units=units,
                output_path=output_path,
                output_format="png",
            )

        if should_plot_all_at_end_fits:

            fits_path = af.path_util.make_and_return_path_from_path_and_folder_names(
                path=output_path, folder_names=["fits"]
            )

            ray_tracing_plotters.plot_ray_tracing_individual(
                tracer=tracer,
                grid=grid,
                mask=mask,
                extract_array_from_mask=extract_array_from_mask,
                zoom_around_mask=zoom_around_mask,
                positions=positions,
                should_plot_profile_image=True,
                should_plot_source_plane=True,
                should_plot_convergence=True,
                should_plot_potential=True,
                should_plot_deflections=True,
                output_path=fits_path,
                output_format="fits",
            )


def plot_lens_fit_for_phase(
    fit,
    during_analysis,
    extract_array_from_mask,
    zoom_around_mask,
    positions,
    units,
    should_plot_mask,
    should_plot_image_plane_pix,
    should_plot_all_at_end_png,
    should_plot_all_at_end_fits,
    should_plot_fit_as_subplot,
    should_plot_fit_of_planes_as_subplot,
    should_plot_inversion_as_subplot,
    should_plot_image,
    should_plot_noise_map,
    should_plot_signal_to_noise_map,
    should_plot_model_image,
    should_plot_residual_map,
    should_plot_normalized_residual_map,
    should_plot_chi_squared_map,
    should_plot_pixelization_residual_map,
    should_plot_pixelization_normalized_residual_map,
    should_plot_pixelization_chi_squared_map,
    should_plot_pixelization_regularization_weights,
    should_plot_subtracted_images_of_planes,
    should_plot_model_images_of_planes,
    should_plot_plane_images_of_planes,
    visualize_path,
    subplot_path,
):

    output_path = af.path_util.make_and_return_path_from_path_and_folder_names(
        path=visualize_path, folder_names=["lens_fit"]
    )

    if should_plot_fit_as_subplot:

        lens_fit_plotters.plot_fit_subplot(
            fit=fit,
            should_plot_mask=should_plot_mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            positions=positions,
            should_plot_image_plane_pix=should_plot_image_plane_pix,
            units=units,
            output_path=subplot_path,
            output_format="png",
        )

    if should_plot_fit_of_planes_as_subplot:

        lens_fit_plotters.plot_fit_subplot_of_planes(
            fit=fit,
            should_plot_mask=should_plot_mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            positions=positions,
            should_plot_image_plane_pix=should_plot_image_plane_pix,
            units=units,
            output_path=subplot_path,
            output_format="png",
        )

    if should_plot_inversion_as_subplot and fit.tracer.has_pixelization:

        inversion_plotters.plot_inversion_subplot(
            inversion=fit.inversion,
            mask=fit.mask_2d,
            positions=positions,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            output_path=subplot_path,
            output_format="png",
        )

    lens_fit_plotters.plot_fit_individuals(
        fit=fit,
        should_plot_mask=should_plot_mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        should_plot_image_plane_pix=should_plot_image_plane_pix,
        should_plot_image=should_plot_image,
        should_plot_noise_map=should_plot_noise_map,
        should_plot_signal_to_noise_map=should_plot_signal_to_noise_map,
        should_plot_model_image=should_plot_model_image,
        should_plot_residual_map=should_plot_residual_map,
        should_plot_chi_squared_map=should_plot_chi_squared_map,
        should_plot_normalized_residual_map=should_plot_normalized_residual_map,
        should_plot_pixelization_residual_map=should_plot_pixelization_residual_map,
        should_plot_pixelization_normalized_residual_map=should_plot_pixelization_normalized_residual_map,
        should_plot_pixelization_chi_squared_map=should_plot_pixelization_chi_squared_map,
        should_plot_pixelization_regularization_weight_map=should_plot_pixelization_regularization_weights,
        should_plot_subtracted_images_of_planes=should_plot_subtracted_images_of_planes,
        should_plot_model_images_of_planes=should_plot_model_images_of_planes,
        should_plot_plane_images_of_planes=should_plot_plane_images_of_planes,
        units=units,
        output_path=output_path,
        output_format="png",
    )

    if not during_analysis:

        if should_plot_all_at_end_png:

            lens_fit_plotters.plot_fit_individuals(
                fit=fit,
                should_plot_mask=should_plot_mask,
                extract_array_from_mask=extract_array_from_mask,
                zoom_around_mask=zoom_around_mask,
                positions=positions,
                should_plot_image_plane_pix=should_plot_image_plane_pix,
                should_plot_image=True,
                should_plot_noise_map=True,
                should_plot_signal_to_noise_map=True,
                should_plot_model_image=True,
                should_plot_residual_map=True,
                should_plot_normalized_residual_map=True,
                should_plot_chi_squared_map=True,
                should_plot_pixelization_residual_map=True,
                should_plot_pixelization_normalized_residual_map=True,
                should_plot_pixelization_chi_squared_map=True,
                should_plot_pixelization_regularization_weight_map=True,
                should_plot_subtracted_images_of_planes=True,
                should_plot_model_images_of_planes=True,
                should_plot_plane_images_of_planes=True,
                units=units,
                output_path=output_path,
                output_format="png",
            )

        if should_plot_all_at_end_fits:

            fits_path = af.path_util.make_and_return_path_from_path_and_folder_names(
                path=output_path, folder_names=["fits"]
            )

            lens_fit_plotters.plot_fit_individuals(
                fit=fit,
                should_plot_mask=should_plot_mask,
                extract_array_from_mask=extract_array_from_mask,
                zoom_around_mask=zoom_around_mask,
                positions=positions,
                should_plot_image_plane_pix=should_plot_image_plane_pix,
                should_plot_image=True,
                should_plot_noise_map=True,
                should_plot_signal_to_noise_map=True,
                should_plot_model_image=True,
                should_plot_residual_map=True,
                should_plot_normalized_residual_map=True,
                should_plot_chi_squared_map=True,
                should_plot_pixelization_residual_map=True,
                should_plot_pixelization_normalized_residual_map=True,
                should_plot_pixelization_chi_squared_map=True,
                should_plot_pixelization_regularization_weight_map=True,
                should_plot_subtracted_images_of_planes=True,
                should_plot_model_images_of_planes=True,
                should_plot_plane_images_of_planes=True,
                output_path=fits_path,
                output_format="fits",
            )


def plot_hyper_images_for_phase(
    hyper_model_image_2d,
    hyper_galaxy_image_2d_path_dict,
    binned_hyper_galaxy_image_2d_path_dict,
    mask,
    binned_grid,
    extract_array_from_mask,
    zoom_around_mask,
    units,
    should_plot_hyper_model_image,
    should_plot_hyper_galaxy_images,
    should_plot_binned_hyper_galaxy_images,
    visualize_path,
):

    output_path = af.path_util.make_and_return_path_from_path_and_folder_names(
        path=visualize_path, folder_names=["hyper_galaxies"]
    )

    if should_plot_hyper_model_image:

        hyper_plotters.plot_hyper_model_image(
            hyper_model_image=hyper_model_image_2d,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            units=units,
            output_path=output_path,
            output_format="png",
        )

    if should_plot_hyper_galaxy_images:

        hyper_plotters.plot_hyper_galaxy_images_subplot(
            hyper_galaxy_image_path_dict=hyper_galaxy_image_2d_path_dict,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            units=units,
            output_path=output_path,
            output_format="png",
        )

    if (
        should_plot_binned_hyper_galaxy_images
        and binned_hyper_galaxy_image_2d_path_dict is not None
    ):

        hyper_plotters.plot_binned_hyper_galaxy_images_subplot(
            hyper_galaxy_cluster_image_path_dict=binned_hyper_galaxy_image_2d_path_dict,
            mask=binned_grid.mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            units=units,
            output_path=output_path,
            output_format="png",
        )



from autolens import decorator_util
import numpy as np

from autolens import exc


class Convolver(object):
    def __init__(self, mask, psf, blurring_mask=None):
        """ Class to setup the 1D convolution of an / mapping_util matrix.

        Take a simple 3x3 and masks:

        [[2, 8, 2],
        [5, 7, 5],
        [3, 1, 4]]

        [[True, False, True],   (True means that the value is masked)
        [False, False, False],
        [True, False, True]]

        A set of values in a corresponding 1d array of this might be represented as:

        [2, 8, 2, 5, 7, 5, 3, 1, 4]

        and after masking as:

        [8, 5, 7, 5, 1]

        Setup is required to perform 2D real-space convolution on the masked array. This module finds the \
        relationship between the unmasked 2D datas, masked datas and psf, so that 2D real-space convolutions \
        can be efficiently applied to reduced 1D masked arrays.

        This calculation also accounts for the blurring of light outside of the masked regions which blurs into \
        the masked region.

        IMAGE FRAMES:
        ------------

        For a masked in 2D, one can compute for every pixel all of the unmasked pixels it will blur light into for \
        a given PSF psf size, e.g.:

        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|     This is an ccd.Mask, where:
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|     x = True (Pixel is masked and excluded from lens)
        |x|x|x|o|o|o|x|x|x|x|     o = False (Pixel is not masked and included in lens)
        |x|x|x|o|o|o|x|x|x|x|
        |x|x|x|o|o|o|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|

        Here, there are 9 unmasked pixels. Indexing of each unmasked pixel goes from the top-left corner right and \
        downwards, therefore:

        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|0|1|2|x|x|x|x|
        |x|x|x|3|4|5|x|x|x|x|
        |x|x|x|6|7|8|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|

        For every unmasked pixel, the Convolver over-lays the PSF and computes three quantities;

        image_frame_indexes - The indexes of all masked pixels it will blur light into.
        image_frame_psfs - The psf values that overlap each masked pixel it will blur light into.
        image_frame_length - The number of masked pixels it will blur light into (unmasked pixels are excluded)

        For example, if we had the following 3x3 psf:

        |0.1|0.2|0.3|
        |0.4|0.5|0.6|
        |0.7|0.8|0.9|

        For pixel 0 above, when we overlap the psf 4 unmasked pixels overlap this psf, such that:

        image_frame_indexes = [0, 1, 3, 4]
        image_frame_psfs = [0.5, 0.6, 0.8, 0.9]
        image_frame_length = 4

        Noting that the other 5 psf values (0.1, 0.2, 0.3, 0.4, 0.7) overlap masked pixels and are thus discarded.

        For pixel 1, we get the following results:

        image_frame_indexes = [0, 1, 2, 3, 4, 5]
        image_frame_psfs = [0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
        image_frame_lengths = 6

        In the majority of cases, the psf will overlap only unmasked pixels. This is the case above for \
        central pixel 4, where:

        image_frame_indexes = [0, 1, 2, 3, 4, 5, 6, 7, 8]
        image_frame_psfs = [0,1, 0.2, 0,3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
        image_frame_lengths = 9

        Once we have set up all these quantities, the convolution routine simply uses them to convolve a 1D array of a
        masked or the masked of a mapping_util in the inversion module.

        BLURRING FRAMES:
        --------------

        Whilst the scheme above accounts for all blurred light within the masks, it does not account for the fact that \
        pixels outside of the masks will also blur light into it. This effect is accounted for using blurring frames.

        It is omitted for mapping_util matrix blurring, as an inversion does not fit datas outside of the masks.

        First, a blurring masks is computed from a masks, which describes all pixels which are close enough to the masks \
        to blur light into it for a given psf size. Following the example above, the following blurring masks is \
        computed:

        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|     This is an example grid.Mask, where:
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|o|o|o|o|o|x|x|x|     x = True (Pixel is masked and excluded from lens)
        |x|x|o|x|x|x|o|x|x|x|     o = False (Pixel is not masked and included in lens)
        |x|x|o|x|x|x|o|x|x|x|
        |x|x|o|x|x|x|o|x|x|x|
        |x|x|o|o|o|o|o|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|

        Indexing again goes from the top-left corner right and downwards:

        |x|x| x| x| x| x| x|x|x|x|
        |x|x| x| x| x| x| x|x|x|x|
        |x|x| x| x| x| x| x|x|x|x|
        |x|x| 0| 1| 2| 3| 4|x|x|x|
        |x|x| 5| x| x| x| 6|x|x|x|
        |x|x| 7| x| x| x| 8|x|x|x|
        |x|x| 9| x| x| x|10|x|x|x|
        |x|x|11|12|13|14|15|x|x|x|
        |x|x| x| x| x| x| x|x|x|x|
        |x|x| x| x| x| x| x|x|x|x|

        For every unmasked blurring-pixel, the Convolver over-lays the PSF psf and computes three quantities;

        blurring_frame_indexes - The indexes of all unmasked pixels (not unmasked blurring pixels) it will \
        blur light into.
        bluring_frame_kernels - The psf values that overlap each pixel it will blur light into.
        blurring_frame_length - The number of pixels it will blur light into.

        The blurring frame therefore does not perform any blurring which blurs light into other blurring pixels. \
        It only performs computations which add light inside of the masks.

        For pixel 0 above, when we overlap the 3x3 psf above only 1 unmasked pixels overlaps the psf, such that:

        blurring_frame_indexes = [0] (This 0 refers to pixel 0 within the masks, not blurring_frame_pixel 0)
        blurring_frame_psfs = [0.9]
        blurring_frame_length = 1

        For pixel 1 above, when we overlap the 3x3 psf above 2 unmasked pixels overlap the psf, such that:

        blurring_frame_indexes = [0, 1]  (This 0 and 1 refer to pixels 0 and 1 within the masks)
        blurring_frame_psfs = [0.8, 0.9]
        blurring_frame_length = 2

        For pixel 3 above, when we overlap the 3x3 psf above 3 unmasked pixels overlap the psf, such that:

        blurring_frame_indexes = [0, 1, 2]  (Again, these are pixels 0, 1 and 2)
        blurring_frame_psfs = [0.7, 0.8, 0.9]
        blurring_frame_length = 3

        Parameters
        ----------
        mask : Mask
            The masks, where True eliminates datas.
        blurring_mask : Mask
            A masks of pixels outside the masks but whose light blurs into it after PSF convolution.
        psf : grid.PSF or ndarray
            An array representing a PSF.
        """

        if psf.shape[0] % 2 == 0 or psf.shape[1] % 2 == 0:
            raise exc.ConvolutionException("PSF kernel must be odd")

        self.mask = mask

        self.mask_index_array = np.full(mask.shape, -1)
        self.pixels_in_mask = int(np.size(mask) - np.sum(mask))

        count = 0
        for x in range(mask.shape[0]):
            for y in range(mask.shape[1]):
                if not mask[x, y]:
                    self.mask_index_array[x, y] = count
                    count += 1

        self.psf = psf
        self.psf_max_size = self.psf.shape[0] * self.psf.shape[1]

        mask_1d_index = 0
        self.image_frame_1d_indexes = np.zeros(
            (self.pixels_in_mask, self.psf_max_size), dtype="int"
        )
        self.image_frame_1d_psfs = np.zeros((self.pixels_in_mask, self.psf_max_size))
        self.image_frame_1d_lengths = np.zeros((self.pixels_in_mask), dtype="int")
        for x in range(self.mask_index_array.shape[0]):
            for y in range(self.mask_index_array.shape[1]):
                if not mask[x][y]:
                    image_frame_1d_indexes, image_frame_1d_psfs = self.frame_at_coordinates_jit(
                        (x, y), mask, self.mask_index_array, self.psf[:, :]
                    )
                    self.image_frame_1d_indexes[
                        mask_1d_index, :
                    ] = image_frame_1d_indexes
                    self.image_frame_1d_psfs[mask_1d_index, :] = image_frame_1d_psfs
                    self.image_frame_1d_lengths[mask_1d_index] = image_frame_1d_indexes[
                        image_frame_1d_indexes >= 0
                    ].shape[0]
                    mask_1d_index += 1

        self.blurring_mask = blurring_mask

        if blurring_mask is not None:

            if mask.shape != blurring_mask.shape:
                raise exc.ConvolutionException(
                    "Mask and Blurring masks must be same shape to generate Convolver"
                )

            self.pixels_in_blurring_mask = int(
                np.size(blurring_mask) - np.sum(blurring_mask)
            )

            mask_1d_index = 0
            self.blurring_frame_1d_indexes = np.zeros(
                (self.pixels_in_blurring_mask, self.psf_max_size), dtype="int"
            )
            self.blurring_frame_1d_psfs = np.zeros(
                (self.pixels_in_blurring_mask, self.psf_max_size)
            )
            self.blurring_frame_1d_lengths = np.zeros(
                (self.pixels_in_blurring_mask), dtype="int"
            )
            for x in range(mask.shape[0]):
                for y in range(mask.shape[1]):
                    if mask[x][y] and not blurring_mask[x, y]:
                        image_frame_1d_indexes, image_frame_1d_psfs = self.frame_at_coordinates_jit(
                            (x, y), mask, self.mask_index_array, self.psf
                        )
                        self.blurring_frame_1d_indexes[
                            mask_1d_index, :
                        ] = image_frame_1d_indexes
                        self.blurring_frame_1d_psfs[
                            mask_1d_index, :
                        ] = image_frame_1d_psfs
                        self.blurring_frame_1d_lengths[
                            mask_1d_index
                        ] = image_frame_1d_indexes[image_frame_1d_indexes >= 0].shape[0]
                        mask_1d_index += 1

    def convolver_with_blurring_mask_added(self, blurring_mask):
        return Convolver(mask=self.mask, psf=self.psf, blurring_mask=blurring_mask)

    @staticmethod
    @decorator_util.jit()
    def frame_at_coordinates_jit(coordinates, mask, mask_index_array, psf):
        """ Compute the frame (indexes of pixels light is blurred into) and psf_frame (psf kernel values of those \
        pixels) for a given coordinate in a masks and its PSF.

        Parameters
        ----------
        coordinates: (int, int)
            The coordinates of mask_index_array on which the frame should be centred
        psf_shape: (int, int)
            The shape of the psf for which this frame will be used
        """

        psf_shape = psf.shape
        psf_max_size = psf_shape[0] * psf_shape[1]

        half_x = int(psf_shape[0] / 2)
        half_y = int(psf_shape[1] / 2)

        frame = -1 * np.ones((psf_max_size))
        psf_frame = -1.0 * np.ones((psf_max_size))

        count = 0
        for i in range(psf_shape[0]):
            for j in range(psf_shape[1]):
                x = coordinates[0] - half_x + i
                y = coordinates[1] - half_y + j
                if (
                    0 <= x < mask_index_array.shape[0]
                    and 0 <= y < mask_index_array.shape[1]
                ):
                    value = mask_index_array[x, y]
                    if value >= 0 and not mask[x, y]:
                        frame[count] = value
                        psf_frame[count] = psf[i, j]
                        count += 1

        return frame, psf_frame

    def convolve_image(self, image_array, blurring_array):
        """For a given 1D array and blurring array, convolve the two using this convolver.

        Parameters
        -----------
        image_array : ndarray
            1D array of the values which are to be blurred with the convolver's PSF.
        blurring_array : ndarray
            1D array of the blurring values which blur into the array after PSF convolution.
        """

        if self.blurring_mask is None:
            raise exc.ConvolutionException(
                "You cannot use the convolve_image function of a Convolver if the Convolver was"
                "not created with a blurring_mask."
            )

        return self.convolve_jit(
            image_1d_array=image_array,
            image_frame_1d_indexes=self.image_frame_1d_indexes,
            image_frame_1d_psfs=self.image_frame_1d_psfs,
            image_frame_1d_lengths=self.image_frame_1d_lengths,
            blurring_1d_array=blurring_array,
            blurring_frame_1d_indexes=self.blurring_frame_1d_indexes,
            blurring_frame_1d_psfs=self.blurring_frame_1d_psfs,
            blurring_frame_1d_lengths=self.blurring_frame_1d_lengths,
        )

    @staticmethod
    @decorator_util.jit()
    def convolve_jit(
        image_1d_array,
        image_frame_1d_indexes,
        image_frame_1d_psfs,
        image_frame_1d_lengths,
        blurring_1d_array,
        blurring_frame_1d_indexes,
        blurring_frame_1d_psfs,
        blurring_frame_1d_lengths,
    ):

        blurred_image_1d = np.zeros(image_1d_array.shape)

        for image_1d_index in range(len(image_1d_array)):

            frame_1d_indexes = image_frame_1d_indexes[image_1d_index]
            frame_1d_psf = image_frame_1d_psfs[image_1d_index]
            frame_1d_length = image_frame_1d_lengths[image_1d_index]
            image_value = image_1d_array[image_1d_index]

            for psf_1d_index in range(frame_1d_length):

                vector_index = frame_1d_indexes[psf_1d_index]
                psf_value = frame_1d_psf[psf_1d_index]
                blurred_image_1d[vector_index] += image_value * psf_value

        for blurring_1d_index in range(len(blurring_1d_array)):

            frame_1d_indexes = blurring_frame_1d_indexes[blurring_1d_index]
            frame_1d_psf = blurring_frame_1d_psfs[blurring_1d_index]
            frame_1d_length = blurring_frame_1d_lengths[blurring_1d_index]
            image_value = blurring_1d_array[blurring_1d_index]

            for psf_1d_index in range(frame_1d_length):
                vector_index = frame_1d_indexes[psf_1d_index]
                psf_value = frame_1d_psf[psf_1d_index]
                blurred_image_1d[vector_index] += image_value * psf_value

        return blurred_image_1d

    def convolve_mapping_matrix(self, mapping_matrix):
        """For a given inversion mapping_util matrix, convolve every pixel's mapped with the PSF kernel.

        A mapping_util matrix provides non-zero entries in all elements which map two pixels to one another
        (see *inversions.mappers*).

        For example, lets take an which is masked using a 'cross' of 5 pixels:

        [[ True, False,  True]],
        [[False, False, False]],
        [[ True, False,  True]]

        As example mapping_util matrix of this cross is as follows (5 pixels x 3 source pixels):

        [1, 0, 0] [0->0]
        [1, 0, 0] [1->0]
        [0, 1, 0] [2->1]
        [0, 1, 0] [3->1]
        [0, 0, 1] [4->2]

        For each source-pixel, we can create an of its unit-surface brightnesses by mapping_util the non-zero
        entries back to masks. For example, doing this for source pixel 1 gives:

        [[0.0, 1.0, 0.0]],
        [[1.0, 0.0, 0.0]]
        [[0.0, 0.0, 0.0]]

        And source pixel 2:

        [[0.0, 0.0, 0.0]],
        [[0.0, 1.0, 1.0]]
        [[0.0, 0.0, 0.0]]

        We then convolve each of these with our PSF kernel, in 2 dimensions, like we would a hyper grid. For
        example, using the kernel below:

        kernel:

        [[0.0, 0.1, 0.0]]
        [[0.1, 0.6, 0.1]]
        [[0.0, 0.1, 0.0]]

        Blurred Source Pixel 1 (we don't need to perform the convolution into masked pixels):

        [[0.0, 0.6, 0.0]],
        [[0.6, 0.0, 0.0]],
        [[0.0, 0.0, 0.0]]

        Blurred Source pixel 2:

        [[0.0, 0.0, 0.0]],
        [[0.0, 0.7, 0.7]],
        [[0.0, 0.0, 0.0]]

        Finally, we map each of these blurred back to a blurred mapping_util matrix, which is analogous to the
        mapping_util matrix.

        [0.6, 0.0, 0.0] [0->0]
        [0.6, 0.0, 0.0] [1->0]
        [0.0, 0.7, 0.0] [2->1]
        [0.0, 0.7, 0.0] [3->1]
        [0.0, 0.0, 0.6] [4->2]

        If the mapping_util matrix is sub-gridded, we perform the convolution on the fractional surface brightnesses in an
        identical fashion to above.

        Parameters
        -----------
        mapping_matrix : ndarray
            The 2D mapping_util matix describing how every inversion pixel maps to an datas_ pixel.
        """
        return self.convolve_matrix_jit(
            mapping_matrix=mapping_matrix,
            image_frame_1d_indexes=self.image_frame_1d_indexes,
            image_frame_1d_psfs=self.image_frame_1d_psfs,
            image_frame_1d_lengths=self.image_frame_1d_lengths,
        )

    @staticmethod
    @decorator_util.jit()
    def convolve_matrix_jit(
        mapping_matrix,
        image_frame_1d_indexes,
        image_frame_1d_psfs,
        image_frame_1d_lengths,
    ):

        blurred_mapping_matrix = np.zeros(mapping_matrix.shape)

        for pixel_1d_index in range(mapping_matrix.shape[1]):
            for image_1d_index in range(mapping_matrix.shape[0]):

                value = mapping_matrix[image_1d_index, pixel_1d_index]

                if value > 0:

                    frame_1d_indexes = image_frame_1d_indexes[image_1d_index]
                    frame_1d_psf = image_frame_1d_psfs[image_1d_index]
                    frame_1d_length = image_frame_1d_lengths[image_1d_index]

                    for psf_1d_index in range(frame_1d_length):
                        vector_index = frame_1d_indexes[psf_1d_index]
                        psf_value = frame_1d_psf[psf_1d_index]
                        blurred_mapping_matrix[vector_index, pixel_1d_index] += (
                            value * psf_value
                        )

        return blurred_mapping_matrix



import autofit as af
import autolens as al

# In this pipeline, we'll perform an analysis which fits an image with the lens light included, and a source galaxy
# using an inversion, using a power-law mass profile. The pipeline follows on from the inversion pipeline
# ''pipelines/with_lens_light/inversion/from_initialize/lens_sersic_sie__source_inversion.py'.

# The pipeline is two phases, as follows:

# Phase 1:

# Description: Fits the lens light and mass model as a power-law, using an inversion for the Source.
# Lens Light: EllipticalSersic
# Lens Mass: EllipitcalPowerLaw + ExternalShear
# Source Light: VoronoiMagnification + Constant
# Previous Pipelines: with_lens_light/inversion/from_initialize/lens_sersic_sie__source_inversion.py
# Prior Passing: Lens Light (variable -> previous pipeline), Lens Mass (variable -> previous pipeline),
#                Source Inversion (variable / constant -> previous pipeline)
# Notes: Uses an interpolation pixel scale for fast power-law deflection angle calculations by default.

# Phase 2:

# Description: Refines the inversion parameters, using a fixed mass model from phase 1.
# Lens Light: EllipticalSersic
# Lens Mass: EllipitcalPowerLaw + ExternalShear
# Source Light: VoronoiMagnification + Constant
# Previous Pipelines: None
# Prior Passing: Lens Mass (constant -> phase 1), source inversion (variable -> phase 1)
# Notes: Uses an interpolation pixel scale for fast power-law deflection angle calculations by default.


def make_pipeline(
    pipeline_settings,
    phase_folders=None,
    redshift_lens=0.5,
    redshift_source=1.0,
    sub_grid_size=2,
    signal_to_noise_limit=None,
    bin_up_factor=None,
    positions_threshold=None,
    inner_mask_radii=None,
    pixel_scale_interpolation_grid=None,
    inversion_uses_border=True,
    inversion_pixel_limit=None,
    pixel_scale_binned_cluster_grid=0.1,
):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline_power_law__lens_sersic_power_law__source_inversion"
    pipeline_tag = al.pipeline_tagging.pipeline_tag_from_pipeline_settings(
        include_shear=pipeline_settings.include_shear,
        fix_lens_light=pipeline_settings.fix_lens_light,
        pixelization=pipeline_settings.pixelization,
        regularization=pipeline_settings.regularization,
    )

    phase_folders.append(pipeline_name)
    phase_folders.append(pipeline_tag)

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's light and mass and one source galaxy, where we:

    # 1) Pass priors on the lens galaxy's light using the EllipticalSersic of the previous pipeline.
    # 2) Pass priors on the lens galaxy's  mass using the EllipticalIsothermal and ExternalShear fit of the previous
    #    pipeline.
    # 3) Pass priors on the source galaxy's inversion using the Pixelization and Regularization of the previous pipeline.

    class LensSourcePhase(al.PhaseImaging):
        def customize_priors(self, results):

            ### Lens Light, Sersic -> Sersic ###

            self.galaxies.lens.light = results.from_phase(
                "phase_2__lens_sersic_sie__source_inversion"
            ).variable.galaxies.lens.light

            self.galaxies.lens.mass = results.from_phase(
                "phase_2__lens_sersic_sie__source_inversion"
            ).variable.galaxies.lens.mass

            self.galaxies.lens.shear = results.from_phase(
                "phase_2__lens_sersic_sie__source_inversion"
            ).variable.galaxies.lens.shear

            ### Lens Mass, SIE -> Powerlaw ###

            self.galaxies.lens.mass.centre = (
                results.from_phase("phase_2__lens_sersic_sie__source_inversion")
                .variable_absolute(a=0.05)
                .galaxies.lens.mass.centre
            )

            self.galaxies.lens.mass.axis_ratio = results.from_phase(
                "phase_2__lens_sersic_sie__source_inversion"
            ).variable.galaxies.lens.mass.axis_ratio

            self.galaxies.lens.mass.phi = results.from_phase(
                "phase_2__lens_sersic_sie__source_inversion"
            ).variable.galaxies.lens.mass.phi

            self.galaxies.lens.mass.einstein_radius = (
                results.from_phase("phase_2__lens_sersic_sie__source_inversion")
                .variable_absolute(a=0.3)
                .galaxies.lens.mass.einstein_radius
            )

            ### Lens Shear, Shear -> Shear ###

            if pipeline_settings.include_shear:

                self.galaxies.lens.shear = results.from_phase(
                    "phase_2__lens_sersic_sie__source_inversion"
                ).variable.galaxies.lens.shear

            ### Source Inversion, Inv -> Inv ###

            self.galaxies.source = results.from_phase(
                "phase_2__lens_sersic_sie__source_inversion"
            ).inversion.constant.galaxies.source

    phase1 = LensSourcePhase(
        phase_name="phase_1__lens_sersic_power_law__source_inversion",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=redshift_lens,
                light=al.light_profiles.EllipticalSersic,
                mass=al.mass_profiles.EllipticalPowerLaw,
                shear=al.mass_profiles.ExternalShear,
            ),
            source=al.GalaxyModel(
                redshift=redshift_source,
                pixelization=pipeline_settings.pixelization,
                regularization=pipeline_settings.regularization,
            ),
        ),
        sub_grid_size=sub_grid_size,
        signal_to_noise_limit=signal_to_noise_limit,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        pixel_scale_interpolation_grid=pixel_scale_interpolation_grid,
        inversion_uses_border=inversion_uses_border,
        inversion_pixel_limit=inversion_pixel_limit,
        pixel_scale_binned_cluster_grid=pixel_scale_binned_cluster_grid,
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 75
    phase1.optimizer.sampling_efficiency = 0.2

    phase1 = phase1.extend_with_inversion_phase()

    return al.PipelineImaging(pipeline_name, phase1)



import sys
import os
import re
import getdist
import numpy as np
from astropy import cosmology

sys.path.append("../autolens/")

import galaxy
from profiles import light_profiles, mass_profiles


def getModelFolders(dir):
    regex = re.compile(".*Hyper")
    folders = [
        name
        for name in os.listdir(dir)
        if os.path.isdir(dir + name) and re.match(regex, name) == None
    ]
    return folders


ltm_skip = 0
center_skip = 0

image_dir = (
    "/gpfs/data_vector/pdtw24/PL_Data/SL03_2/"
)  # Dir of Object to make evidence tables from

image_name = "SLACSJ0252+0039"
redshift = 0.2803
source_redshift = 0.9818
einstein_radius = 1.02
source_light_min = 0.8
source_light_max = 1.3
slacs_ein_r_kpc = 4.18
slacs_mass = "{0:e}".format(10 ** 11.25)
slacs_radius = 4.4
#
# image_name = 'SLACSJ1250+0523'
# redshift = 0.2318
# source_redshift = 0.7953
# einstein_radius = 1.120
# source_light_min = 0.5
# source_light_max = 1.5
# slacs_ein_r_kpc = 4.18
# slacs_mass =  '{0:e}'.format(10**11.26)
# slacs_chab_stellar_mass =  '{0:e}'.format(10**11.53)
# slacs_chab_stellar_mass_lower =  '{0:e}'.format(10**11.46)
# slacs_chab_stellar_mass_upper =  '{0:e}'.format(10**11.60)
# slacs_sal_stellar_mass =  '{0:e}'.format(10**11.77)
# slacs_sal_stellar_mass_lower =  '{0:e}'.format(10**11.7)
# slacs_sal_stellar_mass_upper =  '{0:e}'.format(10**12.84)

image_name = "SLACSJ1430+4105"
redshift = 0.2850
source_redshift = 0.5753
einstein_radius = 1.468
source_light_min = 0.2
source_light_max = 2.3
our_einr_arc = 1.4668492852
slacs_ein_r_kpc = 6.53
slacs_mass = "{0:e}".format(10 ** 11.73)
slacs_chab_stellar_mass = "{0:e}".format(10 ** 11.68)
slacs_chab_stellar_mass_lower = "{0:e}".format(10 ** 11.56)
slacs_chab_stellar_mass_upper = "{0:e}".format(10 ** 11.8)
slacs_sal_stellar_mass = "{0:e}".format(10 ** 11.93)
slacs_sal_stellar_mass_lower = "{0:e}".format(10 ** 11.82)
slacs_sal_stellar_mass_upper = "{0:e}".format(10 ** 12.04)

chab_stellar_frac = 0.33
chab_stellar_frac_error = 0.09

sal_stellar_frac = 0.59
sal_stellar_frac_error = 0.16

# pipeline_folder = 'Pipeline_Total'
# phase_folder = 'PL_Phase_8'

# pipeline_folder = 'Pipeline_LMDM'
# phase_folder = 'PL_Phase_7'

pipeline_folder = "Pipeline_LTM2"
phase_folder = "PL_Phase_7"

image_dir = image_dir + image_name + "/" + pipeline_folder + "/" + phase_folder + "/"

model_folder = getModelFolders(image_dir)

pdf_file = image_dir + model_folder[0] + "/" + image_name + ".txt"

pdf = getdist.mcsamples.loadMCSamples(pdf_file)

values = pdf.getMeans()
params = pdf.getParams()

upper_conf = []
lower_conf = []
for i in range(len(values)):
    upper_conf.append(pdf.confidence(paramVec=i, limfrac=0.95, upper=True))
    lower_conf.append(pdf.confidence(paramVec=i, limfrac=0.95, upper=False))

if image_name == "SLACSJ1250+0523" or image_name == "SLACSJ1430+4105":
    center_skip = 2

if pipeline_folder == "Pipeline_LTM2":
    ltm_skip = 1

sersic_bulge = mass_profiles.SersicMassProfile(
    centre=(values[0], values[1]),
    axis_ratio=values[5],
    phi=values[6],
    intensity=values[2],
    effective_radius=values[3],
    sersic_index=values[4],
    mass_to_light_ratio=values[12 + center_skip],
)

# setattr(sersic_bulge, 'centre_lower_limit_3_sigma', (lower_conf[0], lower_conf[1]))
# setattr(sersic_bulge, 'axis_ratio_lower_limit_3_sigma', lower_conf[5])
# setattr(sersic_bulge, 'phi_lower_limit_3_sigma', lower_conf[6])
# setattr(sersic_bulge, 'intensity_lower_limit_3_sigma', lower_conf[2])
# setattr(sersic_bulge, 'effective_radius_lower_limit_3_sigma', lower_conf[3])
# setattr(sersic_bulge, 'sersic_index_lower_limit_3_sigma', lower_conf[4])
# setattr(sersic_bulge, 'mass_to_light_ratio_lower_limit_3_sigma', lower_conf[12+center_skip])
#
# setattr(sersic_bulge, 'centre_upper_limit_3_sigma', (upper_conf[0], upper_conf[1]))
# setattr(sersic_bulge, 'axis_ratio_upper_limit_3_sigma', upper_conf[5])
# setattr(sersic_bulge, 'phi_upper_limit_3_sigma', upper_conf[6])
# setattr(sersic_bulge, 'intensity_upper_limit_3_sigma', upper_conf[2])
# setattr(sersic_bulge, 'effective_radius_upper_limit_3_sigma', upper_conf[3])
# setattr(sersic_bulge, 'sersic_index_upper_limit_3_sigma', upper_conf[4])
# setattr(sersic_bulge, 'mass_to_light_ratio_upper_limit_3_sigma', upper_conf[12+center_skip])

exponential_halo = mass_profiles.ExponentialMassProfile(
    axis_ratio=values[9 + center_skip],
    phi=values[10 + center_skip],
    intensity=values[7 + center_skip],
    effective_radius=values[8 + center_skip],
    mass_to_light_ratio=values[12 + ltm_skip + center_skip],
)

# setattr(exponential_halo, 'centre_lower_limit_3_sigma', (lower_conf[9+center_skip], lower_conf[10+center_skip]))
# setattr(exponential_halo, 'axis_ratio_lower_limit_3_sigma', lower_conf[9+center_skip])
# setattr(exponential_halo, 'phi_lower_limit_3_sigma', lower_conf[10+center_skip])
# setattr(exponential_halo, 'intensity_lower_limit_3_sigma', lower_conf[7+center_skip])
# setattr(exponential_halo, 'effective_radius_lower_limit_3_sigma', lower_conf[8+center_skip])
# setattr(exponential_halo, 'mass_to_light_ratio_lower_limit_3_sigma', lower_conf[12+ltm_skip+center_skip])
#
# setattr(exponential_halo, 'centre_upper_limit_3_sigma', (upper_conf[9+center_skip], upper_conf[10+center_skip]))
# setattr(exponential_halo, 'axis_ratio_upper_limit_3_sigma', upper_conf[9+center_skip])
# setattr(exponential_halo, 'phi_upper_limit_3_sigma', upper_conf[10+center_skip])
# setattr(exponential_halo, 'intensity_upper_limit_3_sigma', upper_conf[7+center_skip])
# setattr(exponential_halo, 'effective_radius_upper_limit_3_sigma', upper_conf[8+center_skip])
# setattr(exponential_halo, 'mass_to_light_ratio_upper_limit_3_sigma', upper_conf[12+ltm_skip+center_skip])

dark_matter_halo = mass_profiles.SphericalNFWMassProfile(
    kappa_s=values[11 + center_skip], scale_radius=20.0
)

setattr(dark_matter_halo, "kappa_s_lower_limit_3_sigma", lower_conf[11 + center_skip])
setattr(dark_matter_halo, "kappa_s_upper_limit_3_sigma", upper_conf[11 + center_skip])

lens_galaxy = galaxy.Galaxy(
    redshift=redshift, mass_profiles=[sersic_bulge, exponential_halo, dark_matter_halo]
)

lens_galaxy.einstein_radius_in_units = einstein_radius
lens_galaxy.source_light_min = source_light_min
lens_galaxy.source_light_max = source_light_max

source_galaxy = galaxy.Galaxy(redshift=source_redshift)

galaxy.LensingPlanes(
    galaxies=[lens_galaxy, source_galaxy],
    cosmological_model=cosmology.LambdaCDM(H0=70, Om0=0.3, Ode0=0.7),
)

scale_r = 30 * lens_galaxy.arcsec_per_kpc
dark_matter_halo.scale_radius = scale_r
lens_galaxy.mass_profiles[2].scale_radius = scale_r
setattr(dark_matter_halo, "scale_r_lower_limit_3_sigma", scale_r)
setattr(dark_matter_halo, "scale_r_upper_limit_3_sigma", scale_r)

radius = slacs_ein_r_kpc * lens_galaxy.arcsec_per_kpc

total_mass = lens_galaxy.mass_within_circles(radius=radius)
masses = lens_galaxy.mass_within_circles_individual(radius=radius)
stellar_mass = masses[0] + masses[1]

# lens_galaxy.mass_profiles[0].origin = lens_galaxy.mass_profiles[0].centre_lower_limit_3_sigma
# lens_galaxy.mass_profiles[0].axis_ratio = lens_galaxy.mass_profiles[0].axis_ratio_lower_limit_3_sigma
# lens_galaxy.mass_profiles[0].phi = lens_galaxy.mass_profiles[0].phi_lower_limit_3_sigma
# lens_galaxy.mass_profiles[0].intensity = lens_galaxy.mass_profiles[0].intensity_lower_limit_3_sigma
# lens_galaxy.mass_profiles[0].effective_radius = lens_galaxy.mass_profiles[0].effective_radius_lower_limit_3_sigma
# lens_galaxy.mass_profiles[0].sersic_index = lens_galaxy.mass_profiles[0].sersic_index_lower_limit_3_sigma
# lens_galaxy.mass_profiles[0].mass_to_light_ratio = lens_galaxy.mass_profiles[0].mass_to_light_ratio_lower_limit_3_sigma
#
# lens_galaxy.mass_profiles[1].origin = lens_galaxy.mass_profiles[1].centre_lower_limit_3_sigma
# lens_galaxy.mass_profiles[1].axis_ratio = lens_galaxy.mass_profiles[1].axis_ratio_lower_limit_3_sigma
# lens_galaxy.mass_profiles[1].phi = lens_galaxy.mass_profiles[1].phi_lower_limit_3_sigma
# lens_galaxy.mass_profiles[1].intensity = lens_galaxy.mass_profiles[1].intensity_lower_limit_3_sigma
# lens_galaxy.mass_profiles[1].effective_radius = lens_galaxy.mass_profiles[1].effective_radius_lower_limit_3_sigma
# lens_galaxy.mass_profiles[1].mass_to_light_ratio = lens_galaxy.mass_profiles[1].mass_to_light_ratio_lower_limit_3_sigma
#
# lens_galaxy.mass_profiles[2].kappa_s = lens_galaxy.mass_profiles[2].kappa_s_lower_limit_3_sigma
#
# total_mass_lower = lens_galaxy.mass_within_circles(radius=radius)
# masses_lower = lens_galaxy.mass_within_circles_individual(radius=radius)
# stellar_mass_lower = (masses_lower[0] + masses_lower[1])
#
#
# lens_galaxy.mass_profiles[0].origin = lens_galaxy.mass_profiles[0].centre_upper_limit_3_sigma
# lens_galaxy.mass_profiles[0].axis_ratio = lens_galaxy.mass_profiles[0].axis_ratio_upper_limit_3_sigma
# lens_galaxy.mass_profiles[0].phi = lens_galaxy.mass_profiles[0].phi_upper_limit_3_sigma
# lens_galaxy.mass_profiles[0].intensity = lens_galaxy.mass_profiles[0].intensity_upper_limit_3_sigma
# lens_galaxy.mass_profiles[0].effective_radius = lens_galaxy.mass_profiles[0].effective_radius_upper_limit_3_sigma
# lens_galaxy.mass_profiles[0].sersic_index = lens_galaxy.mass_profiles[0].sersic_index_upper_limit_3_sigma
# lens_galaxy.mass_profiles[0].mass_to_light_ratio = lens_galaxy.mass_profiles[0].mass_to_light_ratio_upper_limit_3_sigma
#
# lens_galaxy.mass_profiles[1].origin = lens_galaxy.mass_profiles[1].centre_upper_limit_3_sigma
# lens_galaxy.mass_profiles[1].axis_ratio = lens_galaxy.mass_profiles[1].axis_ratio_upper_limit_3_sigma
# lens_galaxy.mass_profiles[1].phi = lens_galaxy.mass_profiles[1].phi_upper_limit_3_sigma
# lens_galaxy.mass_profiles[1].intensity = lens_galaxy.mass_profiles[1].intensity_upper_limit_3_sigma
# lens_galaxy.mass_profiles[1].effective_radius = lens_galaxy.mass_profiles[1].effective_radius_upper_limit_3_sigma
# lens_galaxy.mass_profiles[1].mass_to_light_ratio = lens_galaxy.mass_profiles[1].mass_to_light_ratio_upper_limit_3_sigma
#
# lens_galaxy.mass_profiles[2].kappa_s = lens_galaxy.mass_profiles[2].kappa_s_upper_limit_3_sigma
#
# total_mass_upper = lens_galaxy.mass_within_circles(radius=radius)
# masses_upper = lens_galaxy.mass_within_circles_individual(radius=radius)
# stellar_mass_upper = (masses_upper[0] + masses_upper[1])

print(masses)

print("einstein radius = ", radius)

masses = lens_galaxy.mass_within_circles_individual(radius=radius)

print()
print("Our Mass = ", "{0:e}".format(total_mass))
# print('Our Mass = ', '{0:e}'.format(total_mass), '(', '{0:e}'.format(total_mass_upper), '{0:e}'.format(total_mass_lower), ')')
print("SLACS Mass = ", slacs_mass)
print("Our Mass / SLACS Mass = ", total_mass / float(slacs_mass))

print()
# print('Stellar Mass fraction in R_Ein = ', stellar_mass / total_mass, '(', stellar_mass_upper / total_mass, stellar_mass_lower / total_mass, ')' )
print("Stellar Mass fraction in R_Ein = ", stellar_mass / total_mass)
print(
    "SLACS Chabrier Stellar Mass Fraction in R_Ein = ",
    chab_stellar_frac,
    " +- ",
    chab_stellar_frac_error,
)
print(
    "SLACS Salpeter Stellar Mass Fraction in R_Ein = ",
    sal_stellar_frac,
    " +- ",
    sal_stellar_frac_error,
)

print()
# zrint('Our Stellar Mass = ','{0:e}'.format((stellar_mass)), '(', '{0:e}'.format(stellar_mass_upper), '{0:e}'.format(stellar_mass_lower), ')' )
print("Our Stellar Mass = ", "{0:e}".format((stellar_mass)))
print(
    "SLACS Chabrier stellar mass = ",
    "{0:e}".format(float(slacs_chab_stellar_mass)),
    "(",
    "{0:e}".format(float(slacs_chab_stellar_mass_lower)),
    "{0:e}".format(float(slacs_chab_stellar_mass_upper)),
    ")",
)
print(
    "SLACS Salpeter stellar mass = ",
    "{0:e}".format(float(slacs_sal_stellar_mass)),
    "(",
    "{0:e}".format(float(slacs_sal_stellar_mass_lower)),
    "{0:e}".format(float(slacs_sal_stellar_mass_upper)),
    ")",
)
print(
    "Our Stellar Mass / SLACS Chabrier Stellar Mass = ",
    stellar_mass / float(slacs_chab_stellar_mass),
)
print(
    "Our Stellar Mass / SLACS Salpeter Stellar Mass = ",
    stellar_mass / float(slacs_sal_stellar_mass),
)


stop

print(
    lens_galaxy.mass_profiles[0].einstein_radius_in_units
    * lens_galaxy.kpc_per_arcsec.value
)
print(
    (
        lens_galaxy.mass_within_circles_individual(
            radius=lens_galaxy.mass_profiles[0].einstein_radius_in_units
            * lens_galaxy.kpc_per_arcsec.value
        )
    )
)
print(lens_galaxy.mass_within_circles_individual(radius=4.18))
print(lens_galaxy.mass_within_circles_individual(radius=radius * 100.0))


stop

lens_galaxy.plot_density_as_function_of_radius(
    maximum_radius=5.0,
    number_bins=300,
    image_name=image_name,
    plot_errors=True,
    labels=[
        "Sersic Bulge",
        "EllipticalExponentialMass Stellar Halo",
        "Dark Matter Halo",
    ],
)



# Welcome to your first pipeline runner, which we'll use to run the tutorial 1 pipeline.
#
# In PyAutoLens, every pipeline is a standalone function which 'makes' a pipeline. We then pass data to this pipeline and
# run it. This keeps our pipelines and data separate, which is good practise, as it it encourages us to write pipelines
# that generalize to as many lenses as possible!

# So, lets begin by discussing the pipeline in this tutorial, which fits the lens and source galaxy of a strong lens.

# In chapter 2, we fitted a strong lens which included the contribution of light from the lens galaxy. We're going to
# fit this lens again (I promise, this is the last time!). However, now we're using pipelines, we can perform a
# completely different (and significantly faster) analysis.

# Load up the PDFs from the previous tutorial -
# 'howtolens/chapter_2_lens_modeling/output/t5_linking_phase_2/image/pdf_triangle.png'.

# This is a big triangle. As we fit models using more and more parameters, its only going to get bigger!

# As usual, you should notice some clear degeneracies between:

# 1) The size (effective_radius, R_l) and intensity (intensity, I_l) of the light profiles.
# 2) The mass normalization (einstein_radius, /Theta_m) and ellipticity (axis_ratio, q_m) of mass profiles.

# This isn't surprising. You can produce similar looking galaxies by trading out intensity for size, and you can
# produce similar mass distributions by compensating for a loss in lens mass by making it a bit less elliptical.

# What do you notice about the contours between the lens galaxy's light-profile and its mass-profile / the source
# galaxy's light profile? Look again.

# That's right - they're not degenerate. The covariance between these sets of parameters is minimal. Again, this makes
# sense - why would fitting the lens's light (which is an elliptical blob of light) be degenerate with fitting the
# source's light (which is a ring of light)? They look nothing like one another!

# So, as a newly trained lens modeler, what does the lack of covariance between these parameters make you think?
# Hopefully, you're thinking, why should I bother fitting the lens and source galaxy simultaneously? Surely we can
# find the right regions of non-linear parameter space by fitting each separately first? This is
# what we're going to do in this tutorial, using a pipeline composed of a modest 3 phases:

# Phase 1) Fit the lens galaxy's light, ignoring the source.

# Phase 2) Fit the source galaxy's light, ignoring the lens.

# Phase 3) Fit both simultaneously, using these results to initialize our starting location in parameter space.

### RUNNER FORMAT ###

# A runner begins by setting up PyAutoFit, in particular the paths to the workspace, config files and where output
# is stored. It is then followed by the setup of PyAutoLens, in particular where data is stored and loaded.

# From here on, we'll use the configs in the 'autolens_workspace/config' folder, which are the default configs used
# by all pipelines (e.g. not just this tutorial, but when you model your own images and lenses!).

# We'll also put the output in 'autolens_workspace/output', which is where output goes for an analysis.

import os

### AUTOFIT + CONFIG SETUP ###

import autofit as af

# Setup the path to the workspace, using by filling in your path below.
workspace_path = "/path/to/user/autolens_workspace/"
workspace_path = "/home/jammy/PycharmProjects/PyAutoLens/workspace/"

# Setup the path to the config folder, using the workspace path.
config_path = workspace_path + "config"

# Use this path to explicitly set the config path and output path.
af.conf.instance = af.conf.Config(
    config_path=config_path, output_path=workspace_path + "output"
)

### AUTOLENS + DATA SETUP ###

import autolens as al

# The same simulate function we used in chapter 2.
def simulate():

    psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.1, pixel_scale=0.1)

    grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
        shape=(130, 130), pixel_scale=0.1, sub_grid_size=2
    )

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.0, 0.0),
            axis_ratio=0.9,
            phi=45.0,
            intensity=0.04,
            effective_radius=0.5,
            sersic_index=3.5,
        ),
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), axis_ratio=0.9, phi=45.0, einstein_radius=1.0
        ),
        shear=al.mass_profiles.ExternalShear(magnitude=0.05, phi=90.0),
    )

    source_galaxy = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.SphericalExponential(
            centre=(0.0, 0.0), intensity=0.2, effective_radius=0.2
        ),
    )

    tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

    return al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
        tracer=tracer,
        grid=grid,
        pixel_scale=0.1,
        exposure_time=300.0,
        psf=psf,
        background_sky_level=0.1,
        add_noise=True,
    )


# Now lets Simulate the CCD data which we'll fit using the pipeline.
ccd_data = simulate()

al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

# To make a pipeline, we call one long function which is written in its own Python module,
# '_tutorial_1_pipeline_lens__source.py_'. Before we check it out, lets get the pipeline running. To do this, we
# import the module and run its 'make_pipeline' function.

# When we run the make_pipeline function, we specify a'phase_folders' which structure the way our output is stored -
# for this pipeline this will output the data as:
# 'autolens_workspace/output/howtolens/c3_t1_lens_and_source/pipeline_name' (the pipeline name is specified in the
# pipeline).
from workspace.howtolens.chapter_3_pipelines import tutorial_1_pipeline_lens_and_source

pipeline_lens_and_source = tutorial_1_pipeline_lens_and_source.make_pipeline(
    phase_folders=["howtolens", "c3_t1_lens_and_source"]
)

# To run a pipeline, we simply use its 'run' function, passing it the data we want to run the pipeline on. Simple, huh?
pipeline_lens_and_source.run(data=ccd_data)

# Okay, good job, we're running our first pipeline in PyAutoLens! But what does it *actually* do? Well, to find that
# out, go to the script '_tutorial_1_pipeline_lens_and_source.py_', which contains a full description of the pipeline,
# as well as an overview of the tools we use to write the most general pipelines possible. Once you're done,
# come back to this pipeline runner script and we'll wrap up tutorial 1.


# And there we have it, a pipeline that breaks the analysis of the lens and source galaxy into 3 simple phases. This
# approach is much faster than fitting the lens and source simultaneously from the beginning. Instead of asking
# you questions at the end of this chapter's tutorials, I'm gonna give a Q&A - this'll hopefully get you thinking about
# how to approach pipeline writing.

# 1) Can this pipeline really be generalized to any lens? Surely the radii of the masks in phase 1 and 2 depend on the
#    lens and source galaxies?
#
#    Whilst this is true, we've chosen values of mask radii above that are 'excessive' and mask out a lot more of the
#    image than just the source (which, in terms of run-time, is desirable). Thus, provided you know the Einstein
#    radius distribution of your lens sample, you can choose masks radii that will masks out every source in your sample
#    adequately (and even if some of the source is still there, who cares? The fit to the lens galaxy will be okay).

#    You can also input custom masks to a pipeline which are used by default if a mask_function is not supplied. I
#    haven't included a howtolens tutorial on this, but if you navigate to the autolens_workspace/pipelines/examples/
#    folder you'll find an example pipeline describingn how to do this ('mask_and_positions.py').

# 2) What if my source galaxy isn't a ring of light? Surely my Annulus masks won't match it?
#
#    Just use the annulus anyway! Yeah, you'll masks out lots of image pixels with no source light, but remember, at
#    the beginning of the pipeline, *we don't care*. In phase 3, we'll use a large circular masks and do the fit
#    properly.



from autolens.model.profiles import light_profiles
from autolens.model.profiles import mass_profiles
from autolens.array import grids
from autolens.model.profiles.plotters import profile_plotters

# In this example, we'll create a grid of Cartesian (y,x) coordinates and pass it to the 'light_profiles'
# module to create images on this grid and the 'mass_profiles' module to create deflection-angle maps on this grid.

# Lets use the same grid as the previous tutorial (if you skipped that tutorial, I recommend you go back to it!)
grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
    shape=(100, 100), pixel_scale=0.05, sub_grid_size=2
)

# Next, lets create a light profile using the 'light_profiles' module. We'll use a Sersic function,
# which is a analytic function often use to depict galaxies.
sersic_light_profile = light_profiles.EllipticalSersic(
    centre=(0.0, 0.0),
    axis_ratio=0.8,
    phi=45.0,
    intensity=1.0,
    effective_radius=1.0,
    sersic_index=2.5,
)

# We can print a profile to confirm its parameters.
print(sersic_light_profile)

# We can pass a grid to a light profile to compute its intensity at every grid coordinate. When we compute an
# array from a grid using a '_from_grid' method like the one below, we have two options for how the calculation is performed:

# 1) The values (e.g. the image) are calculated on the sub-grid. The function will either return the
#    values on this sub grid or bin the sub-gridded values to the grids actual shape (in this case (100, 100)).

# 2) The values are either mapped to the 2D shape of the grid (100, 100) or returned as a flattened 1D NumPy array,
#    on either the sub-grid (e.g. 40000 values) or non sub-gridded array (10000 values).

# This behaviour is determined by two boolean inputs, 'return_in_2d' and 'return_binned'.
light_profile_image = sersic_light_profile.profile_image_from_grid(
    grid=grid, return_in_2d=True, return_binned=True
)

# In this case, the array is returned in 2D and binned up to the shape of the input grid (100, 100).

print("intensity of central grid pixels:")
print(light_profile_image[49, 49])
print(light_profile_image[49, 50])
print(light_profile_image[50, 49])
print(light_profile_image[50, 50])

# We can return the binned-up array in 1D.
light_profile_image = sersic_light_profile.profile_image_from_grid(
    grid=grid, return_in_2d=False, return_binned=True
)

print("intensity of grid pixel 1:")
print(light_profile_image[0])
print("intensity of grid pixel 2:")
print(light_profile_image[1])
print()

# The 1D flattening occurs from the top-left pixel, and goes rightwards and downwards. Thus, because the
# light profile is centered at (0.0, 0.0), the central pixels are the brightest.
print("intensity of central grid pixels:")
print(light_profile_image[4949])
print(light_profile_image[4950])
print(light_profile_image[5049])
print(light_profile_image[5050])

# If we set return_binned to False, we return the values on the sub-grid, which for a sub_grid_size of 2 gives us
# 4 times as many values.

light_profile_image = sersic_light_profile.profile_image_from_grid(
    grid=grid, return_in_2d=False, return_binned=False
)
print("Number of sub pixels = ", light_profile_image.shape[0])

# By default, all arrays from a '_from_grid' method are returned in 2D and binned up, so if you don't specify these
# boolean inputs you'll get the result in the most intuitive format.
light_profile_image = sersic_light_profile.profile_image_from_grid(grid=grid)
print("intensity of central grid pixels:")
print(light_profile_image[49, 49])
print(light_profile_image[49, 50])
print(light_profile_image[50, 49])
print(light_profile_image[50, 50])

# The 'return_in_2d' and 'return_binned' inputs are on pretty much *every* _from_grid methods in PyAutoLens!

# We can use a profile plotter to plot this intensity map (this maps the grid to 2D before plotting).
profile_plotters.plot_image(light_profile=sersic_light_profile, grid=grid)

# To perform ray-tracing, we need to create a 'mass-profile'. A mass-profile is an analytic function that describes the
# distribution of mass in a galaxy, and therefore can be used to derive its surface-density, gravitational potential
# and most importantly, its deflection angles. For those unfamiliar with lensing, the deflection angles describe how
# light is bent by the mass-profile due to the curvature of space-time.

# Lets create a singular isothermal sphere (SIS) mass-profile using the 'mass-profiles' module.
sis_mass_profile = mass_profiles.SphericalIsothermal(
    centre=(0.0, 0.0), einstein_radius=1.6
)

print(sis_mass_profile)

# Just like above, we can pass a grid to a mass-profile to compute its deflection angles (still in 1D)

# (If you are new to gravitiational lensing, and are unclear on what a 'deflection-angle' means or what it is
# used for, then I'll explain all in tutorial 4 of this chapter. For now, just look at the pretty pictures
# they make, and worry about what they mean in tutorial 4!).

mass_profile_deflections = sis_mass_profile.deflections_from_grid(
    grid=grid, return_in_2d=True, return_binned=True
)

print("deflection-angles of grid pixel 0:")
print(mass_profile_deflections[0, 0])
print("deflection-angles of grid pixel 1:")
print(mass_profile_deflections[0, 1])
print()
print("deflection-angles of central grid pixels:")
print(mass_profile_deflections[49, 49])
print(mass_profile_deflections[49, 50])
print(mass_profile_deflections[50, 49])
print(mass_profile_deflections[50, 50])

# And again, a profile plotter can plot these deflection angles in 2D.
profile_plotters.plot_deflections_y(mass_profile=sis_mass_profile, grid=grid)

profile_plotters.plot_deflections_x(mass_profile=sis_mass_profile, grid=grid)

# Mass profiles have a range of other properties that are used for lensing calculations, a couple of which we've
# plotted images of below:

# Convergence - The surface mass density of the mass-profile in dimensionless units which are convenient for lensing calcuations.
# Potential - The gravitational of the mass-profile again in convenient dimensionless units.
# Magnification - Describes how much brighter each image-pixel appears due to focusing of light rays by the mass-profile.

# Extracting arrays of these quantities fom PyAutoLens is exactly the same as for the image and deflection angles above.
mass_profile_convergence = sis_mass_profile.convergence_from_grid(
    grid=grid, return_in_2d=True, return_binned=True
)

mass_profile_potential = sis_mass_profile.potential_from_grid(
    grid=grid, return_in_2d=True, return_binned=True
)

mass_profile_magnification = sis_mass_profile.magnification_from_grid(
    grid=grid, return_in_2d=True, return_binned=True
)

# Plotting them is equally straight forward.

profile_plotters.plot_convergence(mass_profile=sis_mass_profile, grid=grid)

profile_plotters.plot_potential(mass_profile=sis_mass_profile, grid=grid)

profile_plotters.plot_magnification(mass_profile=sis_mass_profile, grid=grid)

# Congratulations, you've completed your second PyAutoLens tutorial! Before moving on to the next one, experiment with
# PyAutoLens by doing the following:
#
# 1) Change the light profile's effective radius and Sersic index - how does the image's appearance change?
# 2) Change the mass profile's einstein radius - what happens to the deflection angles, potential and convergence?
# 3) Experiment with different light-profiles and mass-profiles in the light_profiles and mass_profiles modules.
#    In particular, use the EllipticalIsothermal profile to introduce ellipticity into a mass profile.



import autofit as af
from autolens.array import mask as msk
from autolens.data.plotters import ccd_plotters
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg

import os

# Welcome to the pipeline runner. This tool allows you to load strong lens data, and pass it to pipelines for a
# PyAutoLens analysis. To show you around, we'll load up some example instrument and run it through some of the example
# pipelines that come distributed with PyAutoLens.

# The runner is supplied as both this Python script and a Juypter notebook. Its up to you which you use - I personally
# prefer the python script as provided you keep it relatively small, its quick and easy to comment out different lens
# names and pipelines to perform different analyses. However, notebooks are a tidier way to manage visualization - so
# feel free to use notebooks. Or, use both for a bit, and decide your favourite!

# The pipeline runner is fairly self explanatory. Make sure to checkout the pipelines in the
#  workspace/pipelines/examples/ folder - they come with detailed descriptions of what they do. I hope that you'll
# expand on them for your own personal scientific needs

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))
output_path = workspace_path + "../../outputs/PyAutoLens/slacs_axis_test"

af.conf.instance = af.conf.Config(
    config_path=workspace_path + "config", output_path=output_path
)

# It is convenient to specify the data type and data name as a string, so that if the pipeline is applied to multiple
# images we don't have to change all of the path entries in the load_ccd_data_from_fits function below.

data_type = "slacs"

data_filter = "F814W"

# data_name = 'slacs0216-0813' # Incorrect
data_name = "slacs0252+0039"  # Works
data_name = "slacs0737+3216"  # Works
# data_name = 'slacs0912+0029' # Works
# data_name = 'slacs0959+4410' # Bad fit phase 5 due to source demag thing
# data_name = 'slacs0959+4416' # Bad Fit because its the hard to fit lens
# data_name = 'slacs1011+0143' # x2 lenses
# data_name = 'slacs1205+4910' # Bad fit because foreground galaxy is fitted
# data_name = 'slacs1250+0523' # Bad fit due to poor parametric source.
# data_name = 'slacs1402+6321' # Bad fit and i cant even see the source :/
# data_name = 'slacs1420+6019' # Bad fit phase 5 due to complex source.
# data_name = 'slacs1430+4105' # Works
# data_name = 'slacs1627+0053' # Works, but weird low res source
# data_name = 'slacs1630+4520'
# data_name = 'slacs2238-0754'
# data_name = 'slacs2300+0022'
# data_name = 'slacs2303+1422'

pixel_scale = 0.03

# Create the path where the data will be loaded from, which in this case is
# '/workspace/data/example/lens_light_and_x1_source/'
data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)

from autolens.data.instrument import abstract_data
from autolens.data.instrument import ccd

ccd_data = al.load_ccd_data_from_fits(
    image_path=data_path + "F814W_image.fits",
    psf_path=data_path + "F814W_psf.fits",
    noise_map_path=data_path + "F814W_noise_map.fits",
    pixel_scale=pixel_scale,
    renormalize_psf=True,
    resized_ccd_shape=(301, 301),
    resized_psf_shape=(15, 15),
)

mask = al.load_mask_from_fits(
    mask_path=data_path + "mask.fits", pixel_scale=pixel_scale
)

mask = mask.resized_scaled_array_from_array(new_shape=(301, 301))

positions = al.load_positions(positions_path=data_path + "positions.dat")

ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

# Running a pipeline is easy, we simply import it from the pipelines folder and pass the lens data to its run function.
# Below, we'll' use a 3 phase example pipeline to fit the data with a parametric lens light, mass and source light
# profile. Checkout 'workspace/pipelines/examples/lens_light_and_x1_source_parametric.py' for a full description of
# the pipeline.

# The phase folders input determines the output directory structure of the pipeline, for example the input below makes
# the directory structure:
# 'autolens_workspace/output/phase_folder_1/phase_folder_2/pipeline_name/' or
# 'autolens_workspace/output/example/lens_light_and_x1_source/lens_light_and_x1_source_parametric/'

# For large samples of images, we can therefore easily group lenses that are from the same sample or modeled using the
# same pipeline.

from workspace_jam.pipelines.advanced.with_lens_light.bulge_disk.initialize import (
    lens_bulge_disk_sie__source_sersic,
)
from workspace_jam.pipelines.advanced.with_lens_light.bulge_disk.inversion.from_initialize import (
    lens_bulge_disk_sie__source_inversion,
)

# from workspace_jam.pipelines.with_lens_light.power_law.from_inversion import lens_sersic_power_law__source_inversion

pipeline_pixelization = al.pixelizations.VoronoiBrightnessImage
pipeline_regularization = al.regularization.Constant

pipeline_initialize = lens_bulge_disk_sie__source_sersic.make_pipeline(
    phase_folders=[data_filter, data_name],
    positions_threshold=1.0,
    inner_mask_radii=0.3,
    pipeline_align_bulge_disk_centre=True,
)

pipeline_inversion = lens_bulge_disk_sie__source_inversion.make_pipeline(
    phase_folders=[data_filter, data_name],
    pipeline_fix_lens_light=True,
    pipeline_align_bulge_disk_centre=True,
    pipeline_pixelization=pipeline_pixelization,
    pipeline_regularization=pipeline_regularization,
    positions_threshold=1.0,
    inner_mask_radii=0.3,
)

# pipeline_power_law = lens_sersic_power_law__source_inversion.make_pipeline(phase_folders=[data_type, data_name])

pipeline = (
    pipeline_initialize + pipeline_inversion
)  # + pipeline_power_law + pipeline_subhalo

pipeline.run(data=ccd_data, mask=mask, positions=positions)



import autofit as af
import autolens as al

# In this pipeline, we'll perform a subhalo analysis which determines the sensitivity map of a strong lens and
# then attempts to detection subhalos by putting subhalos at fixed intevals on a 2D (y,x) grid. The source uses a
# light profile. The pipeline is as follows:

# Phase 1:

# Description: Perform the subhalo detection analysis.
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Subhalo: SphericalTruncatedNFWChallenge
# Previous Pipelines: initialize/lens_sie__source_sersic_from_init.py
# Prior Passing: Lens mass (constant -> previous pipeline), source light (variable -> previous pipeline).
# Notes: Priors on subhalo are tuned to give realistic masses (10^6 - 10^10) and concentrations (6-24)

# Phase 2:

# Description: Refine the best-fit detected subhalo from the previous phase, by varying also the lens mass model..
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Subhalo: SphericalTruncatedNFWChallenge
# Previous Pipelines: initialize/lens_sie__source_sersic_from_init.py
# Prior Passing: Lens mass (variable -> previous pipeline), source light and subhalo mass (variable -> phase 2).
# Notes: None


def make_pipeline(
    pipeline_settings,
    phase_folders=None,
    redshift_lens=0.5,
    redshift_source=1.0,
    sub_grid_size=2,
    signal_to_noise_limit=None,
    bin_up_factor=None,
    positions_threshold=None,
    inner_mask_radii=None,
    pixel_scale_interpolation_grid=None,
    parallel=False,
):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline_subhalo_hyper__lens_sie__subhalo_nfw__source_sersic_mass"

    pipeline_tag = al.pipeline_tagging.pipeline_tag_from_pipeline_settings(
        hyper_galaxies=pipeline_settings.hyper_galaxies,
        hyper_image_sky=pipeline_settings.hyper_image_sky,
        hyper_background_noise=pipeline_settings.hyper_background_noise,
        include_shear=pipeline_settings.include_shear,
    )

    phase_folders.append(pipeline_name)
    phase_folders.append(pipeline_tag)

    ### Phase 1 ###

    # In phase 1, we attempt to detect subhalos, by performing a NxN grid search of MultiNest searches, where:

    # 1) The lens model and source-pixelization parameters are held fixed to the best-fit values from phase 1 of the
    #    initialization pipeline.
    # 2) Each grid search varies the subhalo (y,x) coordinates and mass as free parameters.
    # 3) The priors on these (y,x) coordinates are UniformPriors, with limits corresponding to the grid-cells.

    class GridPhase(af.as_grid_search(al.PhaseImaging, parallel=parallel)):
        @property
        def grid_priors(self):
            return [
                self.variable.galaxies.subhalo.mass.centre_0,
                self.variable.galaxies.subhalo.mass.centre_1,
            ]

        def customize_priors(self, results):

            ### Lens Mass, PL -> PL, Shear -> Shear ###

            self.galaxies.lens.mass = results.from_phase(
                "phase_1__lens_sie__source_sersic"
            ).constant.galaxies.lens.mass

            if pipeline_settings.include_shear:

                self.galaxies.lens.shear = results.from_phase(
                    "phase_1__lens_sie__source_sersic"
                ).constant.galaxies.lens.shear

            ### Lens Subhalo, Adjust priors to physical masses (10^6 - 10^10) and concentrations (6-24)

            self.galaxies.subhalo.mass.kappa_s = af.UniformPrior(
                lower_limit=0.0, upper_limit=1.0
            )
            self.galaxies.subhalo.mass.scale_radius = af.UniformPrior(
                lower_limit=0.001, upper_limit=1.0
            )
            self.galaxies.subhalo.mass.centre_0 = af.UniformPrior(
                lower_limit=-2.0, upper_limit=2.0
            )
            self.galaxies.subhalo.mass.centre_1 = af.UniformPrior(
                lower_limit=-2.0, upper_limit=2.0
            )

            ### Source Light, Sersic -> Sersic ###

            self.galaxies.source.light.centre = (
                results.from_phase("phase_1__lens_sie__source_sersic")
                .variable_absolute(a=0.1)
                .galaxies.source.light.centre
            )

            self.galaxies.source.light.intensity = (
                results.from_phase("phase_1__lens_sie__source_sersic")
                .variable_relative(r=0.5)
                .galaxies.source.light.intensity
            )

            self.galaxies.source.light.effective_radius = (
                results.from_phase("phase_1__lens_sie__source_sersic")
                .variable_relative(r=0.5)
                .galaxies.source.light.effective_radius
            )

            self.galaxies.source.light.sersic_index = (
                results.from_phase("phase_1__lens_sie__source_sersic")
                .variable_relative(r=0.5)
                .galaxies.source.light.sersic_index
            )

            self.galaxies.source.light.axis_ratio = (
                results.from_phase("phase_1__lens_sie__source_sersic")
                .variable_absolute(a=0.1)
                .galaxies.source.light.axis_ratio
            )

            self.galaxies.source.light.phi = (
                results.from_phase("phase_1__lens_sie__source_sersic")
                .variable_absolute(a=20.0)
                .galaxies.source.light.phi
            )

            ## Set all hyper_galaxies-galaxies if feature is turned on ##

            if pipeline_settings.hyper_galaxies:

                self.galaxies.source.hyper_galaxy = (
                    results.last.hyper_combined.constant.galaxies.source.hyper_galaxy
                )

            if pipeline_settings.hyper_image_sky:

                self.hyper_image_sky = (
                    results.last.hyper_combined.constant.hyper_image_sky
                )

            if pipeline_settings.hyper_background_noise:

                self.hyper_background_noise = (
                    results.last.hyper_combined.constant.hyper_background_noise
                )

    phase1 = GridPhase(
        phase_name="phase_1__subhalo_search",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=redshift_lens,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=al.mass_profiles.ExternalShear,
            ),
            subhalo=al.GalaxyModel(
                redshift=redshift_lens,
                mass=al.mass_profiles.SphericalTruncatedNFWChallenge,
            ),
            source=al.GalaxyModel(
                redshift=redshift_source, light=al.light_profiles.EllipticalSersic
            ),
        ),
        sub_grid_size=sub_grid_size,
        signal_to_noise_limit=signal_to_noise_limit,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        pixel_scale_interpolation_grid=pixel_scale_interpolation_grid,
        optimizer_class=af.MultiNest,
        number_of_steps=5,
    )

    phase1.optimizer.const_efficiency_mode = False
    phase1.optimizer.n_live_points = 75
    phase1.optimizer.sampling_efficiency = 0.5

    class SubhaloPhase(al.PhaseImaging):
        def customize_priors(self, results):

            ### Lens Mass, PL -> PL, Shear -> Shear ###

            self.galaxies.lens = results.from_phase(
                "phase_1__lens_sie__source_sersic"
            ).variable.galaxies.lens

            ### Source Light, Sersic -> Sersic ###

            self.galaxies.source = results.from_phase(
                "phase_1__subhalo_search"
            ).best_result.variable.galaxies.source

            self.galaxies.source.light.centre = (
                results.from_phase("phase_1__subhalo_search")
                .best_result.variable_absolute(a=0.05)
                .galaxies.source.light.centre
            )

            self.galaxies.source.light.intensity = (
                results.from_phase("phase_1__subhalo_search")
                .best_result.variable_relative(r=0.5)
                .galaxies.source.light.intensity
            )

            self.galaxies.source.light.effective_radius = (
                results.from_phase("phase_1__subhalo_search")
                .best_result.variable_relative(r=0.5)
                .galaxies.source.light.effective_radius
            )

            self.galaxies.source.light.sersic_index = (
                results.from_phase("phase_1__subhalo_search")
                .best_result.variable_relative(r=0.5)
                .galaxies.source.light.sersic_index
            )

            self.galaxies.source.light.axis_ratio = (
                results.from_phase("phase_1__lens_sie__source_sersic")
                .variable_absolute(a=0.1)
                .galaxies.source.light.axis_ratio
            )

            self.galaxies.source.light.phi = (
                results.from_phase("phase_1__lens_sie__source_sersic")
                .variable_absolute(a=10.0)
                .galaxies.source.light.phi
            )

            ## Set all hyper_galaxies-galaxies if feature is turned on ##

            if pipeline_settings.hyper_galaxies:

                self.galaxies.source.hyper_galaxy = results.from_phase(
                    "phase_1__lens_sie__source_sersic"
                ).hyper_combined.constant.galaxies.source.hyper_galaxy

            if pipeline_settings.hyper_image_sky:

                self.hyper_image_sky = results.from_phase(
                    "phase_1__lens_sie__source_sersic"
                ).hyper_combined.constant.galaxies.source.hyper_image_sky

            if pipeline_settings.hyper_background_noise:

                self.hyper_background_noise = results.from_phase(
                    "phase_1__lens_sie__source_sersic"
                ).hyper_combined.constant.galaxies.source.hyper_background_noise

    phase2 = SubhaloPhase(
        phase_name="phase_2__subhalo_refine",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=redshift_lens,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=al.mass_profiles.ExternalShear,
            ),
            subhalo=al.GalaxyModel(
                redshift=redshift_lens,
                mass=phase1.best_result.variable.galaxies.subhalo.mass,
            ),
            source=al.GalaxyModel(
                redshift=redshift_source, light=al.light_profiles.EllipticalSersic
            ),
        ),
        sub_grid_size=sub_grid_size,
        signal_to_noise_limit=signal_to_noise_limit,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        pixel_scale_interpolation_grid=pixel_scale_interpolation_grid,
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 80
    phase2.optimizer.sampling_efficiency = 0.3

    phase2 = phase2.extend_with_multiple_hyper_phases(
        hyper_galaxy=pipeline_settings.hyper_galaxies,
        include_background_sky=pipeline_settings.hyper_image_sky,
        include_background_noise=pipeline_settings.hyper_background_noise,
    )

    return al.PipelineImaging(pipeline_name, phase1, phase2, hyper_mode=True)



import autolens as al
import os

import pytest


@pytest.fixture(name="interferometer_plotter_path")
def make_interferometer_plotter_setup():
    interferometer_plotter_path = "{}/../../test_files/plotting/interferometer/".format(
        os.path.dirname(os.path.realpath(__file__))
    )

    return interferometer_plotter_path


def test__individual_attributes_are_output(
    interferometer_data_7, interferometer_plotter_path, plot_patch
):
    al.interferometer_plotters.plot_visibilities(
        interferometer_data=interferometer_data_7,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=interferometer_plotter_path,
        output_format="png",
    )

    assert (
        interferometer_plotter_path + "interferometer_visibilities.png"
        in plot_patch.paths
    )

    al.interferometer_plotters.plot_u_wavelengths(
        interferometer_data=interferometer_data_7,
        output_path=interferometer_plotter_path,
        output_format="png",
    )

    assert (
        interferometer_plotter_path + "interferometer_u_wavelengths.png"
        in plot_patch.paths
    )

    al.interferometer_plotters.plot_v_wavelengths(
        interferometer_data=interferometer_data_7,
        output_path=interferometer_plotter_path,
        output_format="png",
    )

    assert (
        interferometer_plotter_path + "interferometer_v_wavelengths.png"
        in plot_patch.paths
    )

    al.interferometer_plotters.plot_primary_beam(
        interferometer_data=interferometer_data_7,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=interferometer_plotter_path,
        output_format="png",
    )

    assert (
        interferometer_plotter_path + "interferometer_primary_beam.png"
        in plot_patch.paths
    )

    al.interferometer_plotters.plot_interferometer_subplot(
        interferometer_data=interferometer_data_7,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=interferometer_plotter_path,
        output_format="png",
    )

    assert interferometer_plotter_path + "interferometer_data.png" in plot_patch.paths


def test__interferometer_individuals__output_dependent_on_input(
    interferometer_data_7, general_config, interferometer_plotter_path, plot_patch
):
    al.interferometer_plotters.plot_interferometer_individual(
        interferometer_data=interferometer_data_7,
        should_plot_visibilities=True,
        should_plot_u_wavelengths=False,
        should_plot_v_wavelengths=True,
        should_plot_primary_beam=True,
        output_path=interferometer_plotter_path,
        output_format="png",
    )

    assert (
        interferometer_plotter_path + "interferometer_visibilities.png"
        in plot_patch.paths
    )

    assert (
        not interferometer_plotter_path + "interferometer_u_wavelengths.png"
        in plot_patch.paths
    )

    assert (
        interferometer_plotter_path + "interferometer_v_wavelengths.png"
        in plot_patch.paths
    )

    assert (
        interferometer_plotter_path + "interferometer_primary_beam.png"
        in plot_patch.paths
    )



from autolens.array.grids import reshape_array


class MockLensData(object):
    def __init__(self, ccd_data, mask, grid, blurring_grid, convolver, binned_grid):

        self.ccd_data = ccd_data
        self.unmasked_image = ccd_data.image
        self.unmasked_noise_map = ccd_data.noise_map
        self.pixel_scale = ccd_data.pixel_scale

        self.psf = ccd_data.psf

        self.mask_2d = mask
        self.mask_1d = self.mask_2d.array_1d_from_array_2d(array_2d=self.mask_2d)

        self.grid = grid
        self.grid.new_grid_with_binned_grid(binned_grid=binned_grid)
        self.sub_grid_size = self.grid.sub_grid_size
        self.convolver = convolver

        self.image_1d = self.mask_2d.array_1d_from_array_2d(
            array_2d=self.unmasked_image
        )
        self.noise_map_1d = self.mask_2d.array_1d_from_array_2d(
            array_2d=self.unmasked_noise_map
        )
        self.signal_to_noise_map_1d = self.image_1d / self.noise_map_1d

        self.positions = None

        self.hyper_noise_map_max = None

        self.uses_cluster_inversion = False
        self.inversion_pixel_limit = 1000
        self.inversion_uses_border = True

        self.preload_blurring_grid = blurring_grid
        self.preload_pixelization_grids_of_planes = None

    @reshape_array
    def image(self, return_in_2d=True):
        return self.image_1d

    @reshape_array
    def noise_map(self, return_in_2d=True):
        return self.noise_map_1d

    @reshape_array
    def signal_to_noise_map(self, return_in_2d=True):
        return self.signal_to_noise_map_1d

    @property
    def array_1d_from_array_2d(self):
        return self.grid.array_1d_from_array_2d

    @property
    def scaled_array_2d_from_array_1d(self):
        return self.grid.scaled_array_2d_from_array_1d



import autolens as al
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg

# I think you'll agree, inversions are a very powerful tool for modeling strong lenses. Now that our source galaxies
# comprise just a few parameters we've got a much less complex non-linear parameter space to deal with. This allows us
# to fit more complex mass models and ask ever more interesting scientific questions!

# However, inversions arn't perfect, especially when we use to them model lenses. These arn't huge issues and they're
# easy to deal with, but its worth me explaining them now so they don't trip you up when you start using inversions!

# So, what happens if we fit an image using an inversion and the wrong lens model? lets simulate an image and find out.

# This is the usual simulate function.
def simulate():

    psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
        shape=(180, 180), pixel_scale=0.05
    )

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0, einstein_radius=1.6
        ),
    )

    source_galaxy = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.1, 0.1),
            axis_ratio=0.8,
            phi=90.0,
            intensity=0.2,
            effective_radius=0.3,
            sersic_index=1.0,
        ),
    )

    tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

    return al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
        tracer=tracer,
        grid=grid,
        pixel_scale=0.05,
        exposure_time=300.0,
        psf=psf,
        background_sky_level=0.1,
        add_noise=True,
    )


# And the same fitting function as the last tutorial.
def perform_fit_with_lens__source_galaxy(lens_galaxy, source_galaxy):

    ccd_data = simulate()
    mask = al.Mask.circular_annular(
        shape=ccd_data.shape,
        pixel_scale=ccd_data.pixel_scale,
        inner_radius_arcsec=0.5,
        outer_radius_arcsec=2.2,
    )

    lens_data = al.LensData(ccd_data=ccd_data, mask=mask)

    tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

    return al.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)


# This fit uses a lens galaxy with the wrong mass-model (I've reduced its Einstein Radius from 1.6 to 0.8).
lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0, einstein_radius=0.8
    ),
)

source_galaxy = al.Galaxy(
    redshift=1.0,
    pixelization=pix.Rectangular(shape=(40, 40)),
    regularization=reg.Constant(coefficient=1.0),
)

fit = perform_fit_with_lens__source_galaxy(
    lens_galaxy=lens_galaxy, source_galaxy=source_galaxy
)

al.lens_fit_plotters.plot_fit_subplot(
    fit=fit, should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True
)

# What happened!? This incorrect mass-model provides a really good_fit to the image! The residuals and chi-squared map
# are as good as the ones we saw in the last tutorial.

# How can an incorrect lens model provide such a fit? Well, as I'm sure you noticed, the source has been reconstructed
# as a demagnified version of the image. Clearly, this isn't a physical solution or a solution that we want our
# non-linear search to find, but for inversions these solutions nevertheless exist.

# This isn't necessarily problematic for lens modeling. Afterall, the source reconstruction above is extremely complex,
# in that it requires a lot of pixels to fit the image accurately. Indeed, its Bayesian evidence is much lower than the
# correct solution.
lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0, einstein_radius=1.6
    ),
)

source_galaxy = al.Galaxy(
    redshift=1.0,
    pixelization=pix.Rectangular(shape=(40, 40)),
    regularization=reg.Constant(coefficient=1.0),
)

correct_fit = perform_fit_with_lens__source_galaxy(
    lens_galaxy=lens_galaxy, source_galaxy=source_galaxy
)

al.lens_fit_plotters.plot_fit_subplot(
    fit=correct_fit,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
)

print("Bayesian Evidence of Incorrect Fit:")
print(fit.evidence)
print("Bayesian Evidence of Correct Fit:")
print(correct_fit.evidence)

# The evidence *is* lower. However, the difference in evidence isn't *that large*. This is going to be a problem for our
# non-linear search, as its going to see *a lot* of solutions with really high evidences. Furthermore, these solutions
# occupy the *vast majority* of parameter space (e.g. every single lens model that is wrong). This makes it easy for
# the non-linear search to get lost searching through these unphysical solutions and, unfortunately, infer an incorrect
# lens model (e.g. a local maxima).

# There is no simple fix for this. The reality is that for an inversion these solutions exist. This is why pipelines
# were initially conceived - as they offer a simple solution to this problem. We build a pipelin that begins by modeling
# the source galaxy as a light profile, 'initializing' our lens mass model. Then, when we switch to an inversion in the
# next phase, our mass model starts in the correct regions of parameter space and doesn'tget lost sampling these incorrect solutions.
#
# Its not ideal, but its also not a big problem. Furthermore, light-profiles run faster computationally than inversions,
# so we'd probably want to do this anyway!


# Okay, so we've covered incorrect solutions, lets end by noting that we can model profiles and inversions at the same
# time. We do this when we want to simultaneously fit and subtract the light of a lens galaxy and reconstruct its lensed
# source using an inversion. To do this, all we have to do is give the lens galaxy a light profile.


def simulate_lens_with_light_profile():

    psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
        shape=(180, 180), pixel_scale=0.05
    )

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        light=al.light_profiles.SphericalSersic(
            centre=(0.0, 0.0), intensity=0.2, effective_radius=0.8, sersic_index=4.0
        ),
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0, einstein_radius=1.6
        ),
    )

    source_galaxy = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.1, 0.1),
            axis_ratio=0.8,
            phi=90.0,
            intensity=0.2,
            effective_radius=0.3,
            sersic_index=1.0,
        ),
    )

    tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

    return al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
        tracer=tracer,
        grid=grid,
        pixel_scale=0.05,
        exposure_time=300.0,
        psf=psf,
        background_sky_level=0.1,
        add_noise=True,
    )


# When fitting such an image we now want to include the lens's light in the analysis. First, we update our
# mask to be circular so that it includes the central regions of the image and lens galaxy.
ccd_data = simulate_lens_with_light_profile()

mask = al.Mask.circular(
    shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=2.5
)

# As I said above, performing this fit is the same as usual, we just give the lens galaxy a light profile.
lens_galaxy = al.Galaxy(
    redshift=0.5,
    light=al.light_profiles.SphericalSersic(
        centre=(0.0, 0.0), intensity=0.2, effective_radius=0.8, sersic_index=4.0
    ),
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0, einstein_radius=1.6
    ),
)

# These are all the usual things we do when setting up a fit.
source_galaxy = al.Galaxy(
    redshift=1.0,
    pixelization=pix.Rectangular(shape=(40, 40)),
    regularization=reg.Constant(coefficient=1.0),
)

lens_data = al.LensData(ccd_data=ccd_data, mask=mask)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

# This fit now subtracts the lens galaxy's light from the image and fits the resulting source-only image with the
# inversion. When we plot the image, a new panel on the sub-plot appears showing the model image of the lens galaxy.
fit = al.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)

al.lens_fit_plotters.plot_fit_subplot(
    fit=fit, should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True
)

# Of course if the lens subtraction is rubbish so is our fit, so we can be sure that our lens model wants to fit the
# lens galaxy's light accurately (below, I've increased the lens galaxy intensity from 0.2 to 0.3).
lens_galaxy = al.Galaxy(
    redshift=0.5,
    light=al.light_profiles.SphericalSersic(
        centre=(0.0, 0.0), intensity=0.3, effective_radius=0.8, sersic_index=4.0
    ),
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0, einstein_radius=1.6
    ),
)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

fit = al.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)

al.lens_fit_plotters.plot_fit_subplot(
    fit=fit, should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True
)

# And with that, we're done. Finally, I'll point out a few things about what we've covered to get you thinking about
# the next tutorial on adaption.

# - The unphysical solutions above are clearly problematic. Whilst they have lower Bayesian evidences their existance
#   will still impact our inferred lens model. However, the pixelization's that we used in this chapter are not adapted
#   to the images they are fitting and this means that the correct solutions achieve much lower Bayesian evidence values
#   than is actually possible. Thus, once we've covered adaption, these issues will be resolved!

# - When the lens galaxy's light is subtracted perfectly it leaves no residuals. However, if it isn't subtracted
#   perfectly it does leave residuals and these residuals will be fitted by the inversion. If the residual are
#   significant this is going to mess with our source reconstruction and can lead to some pretty nasty
#   systematics. In the next chapter, we'll learn how our adaptive analysis can prevent this residual fitting.



import autofit as af
import autolens as al

# In this pipeline, we'll perform an initialize analysis which fits an image with a source galaxy and a lens light
# component. This reconstructs the source using a pxielized inversion, and uses the light-profile source fit of a
# previous pipeline. The pipeline is as follows:

# Phase 1:

# Description: initialize the inversion's pixelization and regularization hyper_galaxies-parameters, using a previous lens
#              light and mass model.
# Lens Light: EllipticalSersic + EllipticalExponential
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: VoronoiMagnification + Constant
# Previous Pipelines: initialize/lens_sie__source_sersic_from_init.py
# Prior Passing: Lens Mass (variable -> previous pipeline).
# Notes: Uses the lens subtracted image corresponding to the light model of a previous pipeline.

# Phase 2:

# Description: Refine the lens light and mass model and source inversion.
# Lens Light: EllipticalSersic + EllipticalExponential
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: VoronoiMagnification + Constant
# Previous Pipelines: initialize/lens_sie__source_sersic_from_init.py
# Prior Passing: Lens light and mass (variable -> previous pipeline), source inversion (variable -> phase 1).
# Notes: None

# Phase 3:

# Description: Refine the source inversion using this lens light and mass model.
# Lens Light: EllipticalSersic + EllipticalExponential
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: VoronoiMagnification + Constant
# Previous Pipelines: None
# Prior Passing: Lens light and mass (constant -> phase 2), source inversion (variable -> phase 1 & 2).
# Notes: Source inversion resolution varies.


def make_pipeline(
    pipeline_settings,
    phase_folders=None,
    redshift_lens=0.5,
    redshift_source=1.0,
    sub_grid_size=2,
    signal_to_noise_limit=None,
    bin_up_factor=None,
    positions_threshold=None,
    inner_mask_radii=None,
    pixel_scale_interpolation_grid=None,
    inversion_uses_border=True,
    inversion_pixel_limit=None,
    pixel_scale_binned_cluster_grid=0.1,
):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is tagged according to
    # whether the lens light model is fixed throughout the pipeline and if certain components of the lens light's
    # bulge-disk model are aligned.

    pipeline_name = "pipeline_inv__lens_bulge_disk_sie__source_inversion"

    pipeline_tag = al.pipeline_tagging.pipeline_tag_from_pipeline_settings(
        include_shear=pipeline_settings.include_shear,
        fix_lens_light=pipeline_settings.fix_lens_light,
        pixelization=pipeline_settings.pixelization,
        regularization=pipeline_settings.regularization,
        align_bulge_disk_centre=pipeline_settings.align_bulge_disk_centre,
        align_bulge_disk_axis_ratio=pipeline_settings.align_bulge_disk_axis_ratio,
        disk_as_sersic=pipeline_settings.disk_as_sersic,
        align_bulge_disk_phi=pipeline_settings.align_bulge_disk_phi,
    )

    phase_folders.append(pipeline_name)
    phase_folders.append(pipeline_tag)

    ### PHASE 1 ###

    # In phase 1, we initialize the inversion's resolution and regularization coefficient, where we:

    # 1) Use a lens-subtracted image generated by subtracting model lens galaxy image from phase 3 of the initialize
    #    pipeline.
    # 2) Fix our mass model to the lens galaxy mass-model from phase 3 of the initialize pipeline.
    # 3) Use a circular mask which includes all of the source-galaxy light.

    class InversionPhase(al.PhaseImaging):
        def customize_priors(self, results):

            ## Lens Light & Mass, Bulge -> Bulge, Disk -> Disk, SIE -> SIE, Shear -> Shear ###

            self.galaxies.lens = results.from_phase(
                "phase_4__lens_bulge_disk_sie__source_sersic"
            ).constant.galaxies.lens

    phase1 = InversionPhase(
        phase_name="phase_1__source_inversion_initialization",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=redshift_lens,
                bulge=al.light_profiles.EllipticalSersic,
                disk=al.light_profiles.EllipticalExponential,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=al.mass_profiles.ExternalShear,
            ),
            source=al.GalaxyModel(
                redshift=redshift_source,
                pixelization=pipeline_settings.pixelization,
                regularization=pipeline_settings.regularization,
            ),
        ),
        sub_grid_size=sub_grid_size,
        signal_to_noise_limit=signal_to_noise_limit,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        inversion_uses_border=inversion_uses_border,
        pixel_scale_interpolation_grid=pixel_scale_interpolation_grid,
        inversion_pixel_limit=inversion_pixel_limit,
        pixel_scale_binned_cluster_grid=pixel_scale_binned_cluster_grid,
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 20
    phase1.optimizer.sampling_efficiency = 0.8

    ### PHASE 2 ###

    # In phase 2, we fit the len galaxy light, mass and source galaxy simultaneously, using an inversion. We will:

    # 1) Initialize the priors of the lens galaxy and source galaxy from phase 3 of the previous pipeline and phase 1
    #    of this pipeline.
    # 2) Use a circular mask including both the lens and source galaxy light.

    class InversionPhase(al.PhaseImaging):
        def customize_priors(self, results):

            ## Lens Light & Mass, Bulge -> Bulge, Disk -> Disk, SIE -> SIE, Shear -> Shear ###

            self.galaxies.lens = results.from_phase(
                "phase_4__lens_bulge_disk_sie__source_sersic"
            ).variable.galaxies.lens

            # If the lens light is fixed, over-write the pass prior above to fix the lens light model.

            if pipeline_settings.fix_lens_light:

                self.galaxies.lens.bulge = results.from_phase(
                    "phase_4__lens_bulge_disk_sie__source_sersic"
                ).constant.galaxies.lens.bulge

                self.galaxies.lens.disk = results.from_phase(
                    "phase_4__lens_bulge_disk_sie__source_sersic"
                ).constant.galaxies.lens.disk

    phase2 = InversionPhase(
        phase_name="phase_2__lens_bulge_disk_sie__source_inversion",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=redshift_lens,
                bulge=al.light_profiles.EllipticalSersic,
                disk=al.light_profiles.EllipticalExponential,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=al.mass_profiles.ExternalShear,
            ),
            source=al.GalaxyModel(
                redshift=redshift_source,
                pixelization=phase1.result.constant.galaxies.source.pixelization,
                regularization=phase1.result.constant.galaxies.source.regularization,
            ),
        ),
        sub_grid_size=sub_grid_size,
        signal_to_noise_limit=signal_to_noise_limit,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        pixel_scale_interpolation_grid=pixel_scale_interpolation_grid,
        inversion_uses_border=inversion_uses_border,
        inversion_pixel_limit=inversion_pixel_limit,
        pixel_scale_binned_cluster_grid=pixel_scale_binned_cluster_grid,
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 75
    phase2.optimizer.sampling_efficiency = 0.2

    phase2 = phase2.extend_with_inversion_phase()

    return al.PipelineImaging(pipeline_name, phase1, phase2)



import os
import sys

### AUTOFIT + CONFIG SETUP ###

import autofit as af

cosma_username = "dc-nigh1"
cosma_path = "/cosma6/data/dp004/dc-nigh1/"
cosma_data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    cosma_path, folder_names=["data"]
)
cosma_output_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    cosma_path, folder_names=["output", "slacs_hyper"]
)

workspace_path = "{}/../../../../".format(os.path.dirname(os.path.realpath(__file__)))

# We will use a unique config file for each data-set we model on our super computer. This ensures we can customize the
# analysis of each data-set.
config_path = "{}/../".format(os.path.dirname(os.path.realpath(__file__)))

# Lets now use the above paths to set the config path and output path for our Cosma run.
af.conf.instance = af.conf.Config(
    config_path=config_path + "config", output_path=cosma_output_path
)

### AUTOLENS + DATA SETUP ###

import autolens as al


# Lets take a look at a Cosma batch script, which can be found at 'workspace/runners/cosma/batch/pipeline_runner_cosma'.
# When we submit a PyAutoLens job to Cosma, we submit a 'batch' of jobs, whereby each job will run on one CPU of Cosma.
# Thus, if our lens sample contains, lets say, 17 lenses, we'd submit 17 jobs at the same time where each job applies
# our pipeline to each image.

# The fifth line of this batch script - '#SBATCH --array=1-17' is what species this. Its telling Cosma we're going to
# run 17 jobs, and the id's of those jobs will be numbered from 1 to 17. Infact, these ids are passed to this runner,
# and we'll use them to ensure that each jobs loads a different image. Lets get the cosma array id for our job.
cosma_array_id = int(sys.argv[1])

# Now, I just want to really drive home what the above line is doing. For every job we run on Cosma, the cosma_array_id
# will be different. That is, job 1 will get a cosma_array_id of 1, job 2 will get an id of 2, and so on. This is our
# only unique identifier of every job, thus its our only hope of specifying for each job which image they load!

# Fortunately, we're used to specifying the lens name as a string, so that our pipeline can be applied to multiple
# images with ease. On Cosma, we can apply the same logic, but put these strings in a list such that each Cosma job
# loads a different lens name based on its ID. neat, huh?

data_type = "slacs_jam"

data_filter = "F814W"

data_name = []
data_name.append("")  # Task number beings at 1, so keep index 0 blank
data_name.append("slacs0216-0813")  # Index 1
data_name.append("slacs0252+0039")  # Index 2
data_name.append("slacs0737+3216")  # Index 3
data_name.append("slacs0912+0029")  # Index 4
data_name.append("slacs0959+4410")  # Index 5
data_name.append("slacs0959+4416")  # Index 5
data_name.append("slacs1205+4910")  # Index 6
data_name.append("slacs1250+0523")  # Index 7
data_name.append("slacs1402+6321")  # Index 8
data_name.append("slacs1420+6019")  # Index 9
data_name.append("slacs1430+4105")  # Index 10
data_name.append("slacs1627+0053")  # Index 11
data_name.append("slacs1630+4520")  # Index 12
data_name.append("slacs2238-0754")  # Index 13
data_name.append("slacs2300+0022")  # Index 14
data_name.append("slacs2303+1422")  # Index 15

data_name = data_name[cosma_array_id]

pixel_scale = 0.03  # Make sure your pixel scale is correct!

# We now use the lens_name list to load the image on each job, noting that in this example I'm assuming our lenses are
# on the Cosma instrument directory folder called 'slacs'
data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=cosma_data_path, folder_names=[data_type, data_name]
)

ccd_data = al.load_ccd_data_from_fits(
    image_path=data_path + "F814W_image.fits",
    psf_path=data_path + "F814W_psf.fits",
    noise_map_path=data_path + "F814W_noise_map.fits",
    pixel_scale=pixel_scale,
    resized_ccd_shape=(301, 301),
    resized_psf_shape=(15, 15),
)

mask = al.load_mask_from_fits(
    mask_path=data_path + "mask.fits", pixel_scale=pixel_scale
)
mask = mask.resized_scaled_array_from_array(new_shape=(301, 301))

positions = al.load_positions(positions_path=data_path + "positions.dat")

pipeline_settings = al.PipelineSettingsHyper(
    hyper_galaxies=False,
    hyper_image_sky=False,
    hyper_background_noise=False,
    include_shear=True,
    align_bulge_disk_centre=True,
    disk_as_sersic=True,
    fix_lens_light=True,
    pixelization=al.pixelizations.VoronoiBrightnessImage,
    regularization=al.regularization.AdaptiveBrightness,
)

# Running a pipeline is exactly the same as we're used to. We import it, make it, and run it, noting that we can
# use the lens_name's to ensure each job outputs its results to a different directory.

from workspace_jam.pipelines.hyper.with_lens_light.bulge_disk.initialize import (
    lens_bulge_disk_sie__source_sersic,
)
from workspace_jam.pipelines.hyper.with_lens_light.bulge_disk.inversion.from_initialize import (
    lens_bulge_disk_sie__source_inversion,
)
from workspace_jam.pipelines.hyper.with_lens_light.bulge_disk.power_law.from_inversion import (
    lens_bulge_disk_power_law__source_inversion,
)

pipeline_initialize = lens_bulge_disk_sie__source_sersic.make_pipeline(
    pipeline_settings=pipeline_settings,
    phase_folders=[data_filter, data_name],
    positions_threshold=1.0,
)

pipeline_inversion = lens_bulge_disk_sie__source_inversion.make_pipeline(
    pipeline_settings=pipeline_settings,
    phase_folders=[data_filter, data_name],
    positions_threshold=1.0,
)

pipeline_power_law = lens_bulge_disk_power_law__source_inversion.make_pipeline(
    pipeline_settings=pipeline_settings,
    phase_folders=[data_filter, data_name],
    positions_threshold=0.5,
)

pipeline = pipeline_initialize + pipeline_inversion + pipeline_power_law

pipeline.run(data=ccd_data, mask=mask, positions=positions)



from autolens.model.profiles import light_profiles
from autolens.model.profiles import mass_profiles
from autolens.model.galaxy import galaxy
from autolens.lens import ray_tracing
from autolens.array import grids
from autolens.lens.plotters import plane_plotters
from autolens.lens.plotters import ray_tracing_plotters

# In the last tutorial, our use of planes was a bit clunky. We manually had to input grids to trace them, and keep track
# of which grids were image-plane grids and which were source plane grids. It was easy to make mistakes!
#
# Fotunately, in PyAutoLens, you won't actually spend much hands-on time with the plane module. Instead, you'll primarily
# use the 'ray-tracing' module, which we'll cover in this example. Lets look at how easy it is to setup the same
# lens-plane + source-plane strong lens configuration as the previous tutorial, but with a lot less lines of code!

# Let use the same grid we've all grown to know and love by now!
image_plane_grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
    shape=(100, 100), pixel_scale=0.05, sub_grid_size=2
)

# For our lens galaxy, we'll use the same SIS mass profile as before.
sis_mass_profile = mass_profiles.SphericalIsothermal(
    centre=(0.0, 0.0), einstein_radius=1.6
)

lens_galaxy = galaxy.Galaxy(redshift=0.5, mass=sis_mass_profile)
print(lens_galaxy)

# And for our source galaxy, the same Sersic light profile
sersic_light_profile = light_profiles.SphericalSersic(
    centre=(0.0, 0.0), intensity=1.0, effective_radius=1.0, sersic_index=1.0
)

source_galaxy = galaxy.Galaxy(redshift=1.0, light=sersic_light_profile)

print(source_galaxy)

# Now, lets use the lens and source galaxies to ray-trace our grid, using a 'tracer' from the ray-tracing
# module. When we pass our galaxies into the Tracer below, the following happens:

# 1) The galaxies are ordered in ascending redshift.
# 2) Planes are created at every one of these redshifts, with the galaxies at those redshifts associated with those
#    planes.
tracer = ray_tracing.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

# This tracer is composed of a list of planes, in this case two planes (the image and source plane).
print(tracer.planes)

# We can access these using the 'image-plane' and 'source-plane' attributes.
print("Image Plane:")
print(tracer.planes[0])
print(tracer.image_plane)
print()
print("Source Plane:")
print(tracer.planes[1])
print(tracer.source_plane)

# The most convenient part of the tracer is we can use it to perform fully 'ray-traced' images, without manually
# setting up the planes to do this. The function below does the following

# 1) Using the lens-galaxy's mass-profile, the deflection angle of every image-plane grid coordinate is computed.
# 2) These deflection angles are used to trace every image-plane coordinate to a source-plane coordinate.
# 3) The light of each traced source-plane coordinate is evaluated using the source-plane galaxy's light profile.

traced_profile_image = tracer.profile_image_from_grid(grid=image_plane_grid)
print("traced image pixel 1")
print(traced_profile_image[0, 0])
print("traced image pixel 2")
print(traced_profile_image[0, 1])
print("traced image pixel 3")
print(traced_profile_image[0, 2])

# This image appears as the Einstein ring we saw in the previous tutorial.
ray_tracing_plotters.plot_profile_image(tracer=tracer, grid=image_plane_grid)

# We can also use the tracer to compute the traced grid of every plane, instead of getting the traced image itself:
traced_grids = tracer.traced_grids_of_planes_from_grid(
    grid=image_plane_grid, return_in_2d=False
)

# The tracer is composed of an image-plane and source-plane, just like in the previous example!
print("grid image-plane coordinate 1")
print(traced_grids[0][0])
print("grid image-plane coordinate 2")
print(traced_grids[0][1])
print("grid image-plane coordinate 3")
print(traced_grids[0][2])

# And the source-plane's grid has been deflected.
print("grid source-plane coordinate 1")
print(traced_grids[1][0])
print("grid source-plane coordinate 2")
print(traced_grids[1][1])
print("grid source-plane coordinate 3")
print(traced_grids[1][2])

# If we want, we can use the plane_plotters to plot these grids like before.
plane_plotters.plot_plane_grid(
    plane=tracer.image_plane, grid=traced_grids[0], title="Image-plane Grid"
)

plane_plotters.plot_plane_grid(
    plane=tracer.source_plane, grid=traced_grids[1], title="Source-plane Grid"
)

plane_plotters.plot_plane_grid(
    plane=tracer.source_plane,
    grid=traced_grids[1],
    axis_limits=[-0.1, 0.1, -0.1, 0.1],
    title="Source-plane Grid",
)

# As discussed above, interacting directly with planes is a bit comberson. PyAutoLens has tools for plotting
# everything in  a tracer automatically. For example, a ray-tracing subplot plots the following:

# 1) The image, computed by tracing the source galaxy's light 'forwards' through the tracer.
# 2) The source-plane image, showing the source galaxy's true appearance (i.e. if it were not lensed).
# 3) The image-plane convergence, computed using the lens galaxy's mass profile.
# 4) The image-plane gravitational potential, computed using the lens galaxy's mass profile.
# 5) The image-plane deflection angles, computed using the lens galaxy's mass profile.

ray_tracing_plotters.plot_ray_tracing_subplot(tracer=tracer, grid=image_plane_grid)

# Just like for a plane, these quantities attributes can be computed by passing a grid (converted to 2D NumPy
# arrays the same dimensions as our input grid!).

convergence = tracer.convergence_from_grid(
    grid=image_plane_grid, return_in_2d=True, return_binned=True
)

print("Tracer - Convergence - grid coordinate 1:")
print(convergence[0, 0])
print("Tracer - Convergence - grid coordinate 2:")
print(convergence[0, 1])
print("Tracer - Convergence - grid coordinate 3:")
print(convergence[0, 2])
print("Tracer - Convergence - grid coordinate 101:")
print(convergence[1, 0])

# Of course, these convergence are identical to the image-plane convergences, as it's only the lens galaxy
# that contributes to the overall mass of the ray-tracing system.
image_plane_convergence = tracer.image_plane.convergence_from_grid(
    grid=image_plane_grid, return_in_2d=True, return_binned=True
)

print("Image-Plane - Convergence - grid coordinate 1:")
print(image_plane_convergence[0, 0])
print("Image-Plane - Convergence - grid coordinate 2:")
print(image_plane_convergence[0, 1])
print("Image-Plane - Convergence - grid coordinate 3:")
print(image_plane_convergence[0, 2])
print("Image-Plane - Convergence - grid coordinate 101:")
print(image_plane_convergence[1, 0])

# I've left the rest below commented to avoid too many print statements, but if you're feeling adventurous go ahead
# and uncomment the lines below!
# print('Potential:')
# print(tracer.potential_from_grid(grid=image_plane_grid))
# print(tracer.image_plane.potential_from_grid(grid=image_plane_grid))
# print('Deflections:')
# print(tracer.deflections_from_grid(grid=image_plane_grid))
# print(tracer.deflections_from_grid(grid=image_plane_grid))
# print(tracer.image_plane.deflections_from_grid(grid=image_plane_grid))
# print(tracer.image_plane.deflections_from_grid(grid=image_plane_grid))

# You can also plot the above attributes on individual figures, using appropriate ray-tracing plotter (I've left most
# commented out again for convinience)
ray_tracing_plotters.plot_convergence(tracer=tracer, grid=image_plane_grid)

# ray_tracing_plotters.plot_potential(tracer=tracer, grid=image_plane_grid)
# ray_tracing_plotters.plot_deflections_y(tracer=tracer, grid=image_plane_grid)
# ray_tracing_plotters.plot_deflections_x(tracer=tracer, grid=image_plane_grid)
# ray_tracing_plotters.plot_image_plane_image(tracer=tracer, grid=image_plane_grid)

# Before we finish, you might be wondering 'why do both the image-plane and tracer have the attributes convergence
# / potential / deflection angles, when the two are identical'. Afterall, only mass profiles contribute to these
# quantities, and only the image-plane has galaxies with measureable mass profiles! There are two reasons:

# 1) Convinience - You could always write 'tracer.image_plane.convergence_from_grid' and
#                  'plane_plotters.convergence(plane=tracer.image_plane). However, code appears neater if you can
#                  just write 'tracer.convergence_from_grid' and 'ray_tracing_plotters.plot_convergence(tracer=tracer).

# 2) Multi-plane lensing - For now, we're focused on the simplest lensing configuration possible, an image-plane + source-plane
#                          configuration. However, there are strong lens system where there are more than 2 planes! In these
#                          instances, the convergence, potential and deflections of each plane is different to the overall
#                          values given by the tracer.

#                          This is beyond the scope of this chapter, but be reassured that what you're learning now
#                          will prepare you for the advanced chapters later on!

# And with that, we're done. You've performed your first ray-tracing with PyAutoLens! There are
# no exercises for this chapter, and we're going to take a deeper look at ray-tracing in the next
# chapter.



import autofit as af
import autolens as al

import os

# In this example, we will demonstrate how the appearance of figures in PyAutoLens can be customized. To do this, we
# will use the the image of the strong lens slacs1430+4105 from a .fits file and plot it using the
# function autolens.instrument.plotters.array_plotters.plot_array.

# The customization functions demonstrated in this example are generic to any 2D array of instrument, and can therefore be
# applied to the plotting of noise-maps, PSF's, residual maps, chi-squared maps, etc.

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../../".format(os.path.dirname(os.path.realpath(__file__)))

# We have included the .fits instrument required for this example in the directory
# 'workspace/output/data/example/slacs1430+4105/'.

# First, lets setup the path to the .fits file of the image.
data_type = "example"
data_name = "slacs1430+4105"

# Create the path where the data will be loaded from, which in this case is
# '/workspace/data/example/slacs1430+4105/'
data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)
image_path = data_path + "image.fits"

# Now, lets load this array as a hyper array. A hyper array is an ordinary NumPy array, but it also includes a pixel
# scale which allows us to convert the axes of the array to arc-second coordinates.
image = al.ScaledSquarePixelArray.from_fits_with_pixel_scale(
    file_path=image_path, hdu=0, pixel_scale=0.03
)

# We can now use an array plotter to plot the array. We customize the plot as follows:

# 1) We make the array's figure size bigger than the default size (7,7).

# 2) Because the figure is bigger, we increase the size of the title, x and y labels / ticks from their default size of
#    16 to 24.

# 3) For the same reason, we increase the size of the colorbar ticks from the default value 10 to 20.
al.array_plotters.plot_array(
    array=image,
    figsize=(12, 12),
    title="SLACS1430+4105 Image",
    titlesize=24,
    xlabelsize=24,
    ylabelsize=24,
    xyticksize=24,
    cb_ticksize=20,
)

# The colormap of the array can be changed to any of the standard matplotlib colormaps.
al.array_plotters.plot_array(array=image, title="SLACS1430+4105 Image", cmap="spring")

# We can change the x / y axis units from arc-seconds to kiloparsec, by inputting a kiloparsec to arcsecond conversion
# factor (for SLACS1430+4105, the lens galaxy is at redshift 0.285, corresponding to the conversion factor below).
al.array_plotters.plot_array(
    array=image, title="SLACS1430+4105 Image", units="kpc", kpc_per_arcsec=4.335
)

# The matplotlib figure can be output to the hard-disk as a png, as follows.
al.array_plotters.plot_array(
    array=image,
    title="SLACS1430+4105 Image",
    output_path=workspace_path + "/plotting/plots/",
    output_filename="array",
    output_format="png",
)

# The array itself can be output to the hard-disk as a .fits file.
al.array_plotters.plot_array(
    array=image,
    output_path=workspace_path + "/plotting/plots/",
    output_filename="array",
    output_format="fits",
)



from .phase import GalaxyFitPhase, Phase
from autolens.pipeline.phase.phase_extensions import HyperGalaxyPhase
from .phase_imaging import PhaseImaging, PhaseImaging, PhaseImaging, PhaseImaging
from .phase_positions import PhasePositions



import autofit as af
import autolens as al
import os

# This tool allows one to make simulated data-sets of strong lenses using a multi-plane ray-tracer, such that all
# galaxies down the line-of-sight are included in the ray-tracing calculation based on their redshifts.

# The 'data name' is the name of the data folder and 'data_name' the folder the data is stored in, e.g:

# The image will be output as '/workspace/data/data_type/data_name/image.fits'.
# The noise-map will be output as '/workspace/data/data_type/data_name/lens_name/noise_map.fits'.
# The psf will be output as '/workspace/data/data_type/data_name/psf.fits'.

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../../".format(os.path.dirname(os.path.realpath(__file__)))

# (these files are already in the workspace and are remade running this script)
data_type = "example"
data_name = "multi_plane"

# Create the path where the data will be output, which in this case is
# '/workspace/data/example/multi_plane/'
data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)

# The pixel scale of the image to be simulated
pixel_scale = 0.05

# Simulate a simple Gaussian PSF for the image.
psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=pixel_scale)

# Setup the image-plane al.ogrid of the CCD array which will be used for generating the image of the
# simulated strong lens.
grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
    shape=(400, 400), pixel_scale=pixel_scale, sub_grid_size=1
)

# Setup the lens galaxy's light mass (SIE) and source galaxy light (elliptical Sersic) for this simulated lens.
lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
    ),
    shear=al.mass_profiles.ExternalShear(magnitude=0.05, phi=90.0),
)

source_galaxy = al.Galaxy(
    redshift=1.0,
    light=al.light_profiles.EllipticalSersic(
        centre=(0.1, 0.1),
        axis_ratio=0.8,
        phi=60.0,
        intensity=1.0,
        effective_radius=1.0,
        sersic_index=2.5,
    ),
)

# Setup our line-of-sight (los) galaxies using Spherical Sersic profiles for their light and Singular
# Isothermal Sphere (SIS) profiles. We'll use 3 galaxies, but you can add more if desired.
los_0 = al.Galaxy(
    redshift=0.25,
    light=al.light_profiles.SphericalSersic(
        centre=(4.0, 4.0), intensity=0.30, effective_radius=0.3, sersic_index=2.0
    ),
    mass=al.mass_profiles.SphericalIsothermal(centre=(4.0, 4.0), einstein_radius=0.02),
)

los_1 = al.Galaxy(
    redshift=0.75,
    light=al.light_profiles.SphericalSersic(
        centre=(3.6, -5.3), intensity=0.20, effective_radius=0.6, sersic_index=1.5
    ),
    mass=al.mass_profiles.SphericalIsothermal(centre=(3.6, -5.3), einstein_radius=0.04),
)

los_2 = al.Galaxy(
    redshift=1.25,
    light=al.light_profiles.SphericalSersic(
        centre=(-3.1, -2.4), intensity=0.35, effective_radius=0.4, sersic_index=2.5
    ),
    mass=al.mass_profiles.SphericalIsothermal(
        centre=(-3.1, -2.4), einstein_radius=0.03
    ),
)

# Use these galaxies to setup a multi-plane tracer, which will generate the image for the simulated CCD
# instrument. This tracer orders galaxies by redshift and performs ray-tracing based on their line-of-sight redshifts.
tracer = al.Tracer.from_galaxies(
    galaxies=[lens_galaxy, source_galaxy, los_0, los_1, los_2]
)

# Lets look at the tracer's image - this is the image we'll be simulating.
al.ray_tracing_plotters.plot_ray_tracing_subplot(tracer=tracer, grid=grid)

# Now lets simulate the CCD instrument, remembering that we use a special image which ensures edge-effects don't
# degrade our modeling of the telescope optics (e.g. the PSF convolution).
ccd = al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
    tracer=tracer,
    grid=grid,
    pixel_scale=pixel_scale,
    exposure_time=300.0,
    psf=psf,
    background_sky_level=0.1,
    add_noise=True,
)

# Lets plot the simulated CCD instrument before we output it to files.
al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd)

# Finally, lets output our simulated instrument to the data path as .fits files.
al.output_ccd_data_to_fits(
    ccd_data=ccd,
    image_path=data_path + "image.fits",
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    overwrite=True,
)
